# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import logging
import pathlib
import typing as t

import click

import capellambse
from capellambse import _diagram_cache, cli_helpers

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "-m",
    "--model",
    "modelinfo",
    type=cli_helpers.ModelInfoCLI(),
    required=True,
    help="Capella model to export diagrams from",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(file_okay=False, dir_okay=True, path_type=pathlib.Path),
    default="./diagrams",
    help="Directory to store the rendered diagrams in",
    show_default=True,
    envvar="DIAGRAM_CACHE_OUTPUT_DIR",
    show_envvar=True,
)
@click.option(
    "-f",
    "--format",
    type=click.Choice(sorted(_diagram_cache.VALID_FORMATS)),
    default="svg",
    help="Image output format",
    show_default=True,
    envvar="DIAGRAM_CACHE_OUTPUT_FORMAT",
    show_envvar=True,
)
@click.option(
    "--index/--no-index",
    default=True,
    help="Generate index.json and index.html files",
    show_default=True,
    envvar="DIAGRAM_CACHE_GENERATE_INDEX",
    show_envvar=True,
)
@click.option(
    "--refresh / --no-refresh",
    default=False,
    help=(
        "Refresh all diagrams before exporting them, to make sure their"
        " contents are up to date with the model. Recommended, but"
        " disabled by default because it may take a long time."
        " NOTE: This may temporarily open a Capella window (see above)."
    ),
    envvar="DIAGRAM_CACHE_REFRESH_DIAGRAMS",
    show_envvar=True,
)
@click.option("--exe", help="Name or path of the Capella executable")
@click.option("--docker", help="Name of a Docker image containing Capella")
@click.option(
    "--background/--no-background",
    help="Inserts a white background into the diagrams.",
    default=True,
    show_default=True,
    envvar="DIAGRAM_CACHE_INSERT_BACKGROUND",
    show_envvar=True,
)
def main(
    modelinfo: dict[str, t.Any],
    output: pathlib.Path,
    format: str,
    index: bool,
    exe: str | None,
    docker: str | None,
    background: bool,
    refresh: bool,
) -> None:
    """Export diagrams from a Capella model.

    This tool can be used to easily populate a diagram cache for use
    with `capellambse.MelodyModel`.

    \b
    Exit codes
    ----------

    The CLI will indicate the success status via exit codes:

    \b
    - 0 in case of success
    - 1 for general errors
    - 2 for CLI usage errors
    - 3 if some diagrams failed to export, but others were successful

    \b
    Refreshing representations
    --------------------------

    By default, Capella will only automatically refresh diagrams on
    an as-needed basis whenever a diagram is opened in the GUI. This
    can lead to out-of-date images being written by its
    "exportRepresentations" facility, which this tool uses.

    To avoid this problem, the '--refresh' switch can be used to
    automatically refresh all diagrams and make sure they are up to
    date with the model contents.

    Unfortunately, due to the way this refresh operation is
    implemented in Capella, this may cause a Capella window to be
    opened temporarily. It will automatically close once the
    operation is done.

    PLEASE DO NOT CLOSE THE WINDOW that Capella opens during this
    step, or the entire process will be aborted.

    Also be aware that using this option may lead to new objects
    being added in odd places, which may cause severe readability
    issues on the exported diagrams.

    Due to these issues, the automatic refresh is disabled by
    default.
    """  # noqa: D301
    modelinfo["diagram_cache"] = {"path": output}
    model_ = capellambse.MelodyModel(**modelinfo)

    if not model_.diagrams:
        logger.info("No diagrams found in the model, nothing to export")
        raise SystemExit(0)

    if docker:
        capella = docker
        force = "docker"
    else:
        capella = exe or "capella{VERSION}"
        force = "exe"

    diagrams = _diagram_cache.export(
        capella,
        model_,
        format=format,
        index=index,
        force=force,  # type: ignore
        background=background,
        refresh=refresh,
    )

    ok = sum(1 for i in diagrams if i["success"])
    if ok == 0:
        logger.error("Could not export any diagrams")
        raise SystemExit(1)

    failed = len(diagrams) - ok
    if failed > 0:
        msg = "\n".join(
            f" - {i['name']!r} ({i['uuid']})"
            for i in diagrams
            if not i["success"]
        )
        logger.error(
            "%d diagrams failed to export (%d ok)\n%s", failed, ok, msg
        )
        raise SystemExit(3)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Physical Architecture layer."""

from __future__ import annotations

import enum
import sys
import typing as t

import capellambse.model as m

from .. import capellacommon, cs, fa, information
from .. import namespaces as ns

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

NS = ns.PA


@m.stringy_enum
@enum.unique
class PhysicalComponentKind(enum.Enum):
    """Categories of physical components.

    Allows to categorize a physical component, with respect to real life
    physical entities.
    """

    UNSET = "UNSET"
    """The physical component kind is not specified."""
    HARDWARE = "HARDWARE"
    """The physical component is a hardware resource."""
    HARDWARE_COMPUTER = "HARDWARE_COMPUTER"
    """The physical component is a computing resource."""
    SOFTWARE = "SOFTWARE"
    """The physical component is a software entity."""
    SOFTWARE_DEPLOYMENT_UNIT = "SOFTWARE_DEPLOYMENT_UNIT"
    """The physical component is a software deployment unit."""
    SOFTWARE_EXECUTION_UNIT = "SOFTWARE_EXECUTION_UNIT"
    """The physical component is a software execution unit."""
    SOFTWARE_APPLICATION = "SOFTWARE_APPLICATION"
    """The physical component is a software application."""
    FIRMWARE = "FIRMWARE"
    """The physical component is a firmware part."""
    PERSON = "PERSON"
    """The physical component is a person."""
    FACILITIES = "FACILITIES"
    """The physical component refers to Facilities."""
    DATA = "DATA"
    """The physical component represents a set of data."""
    MATERIALS = "MATERIALS"
    """The physical component represents a bunch of materials."""
    SERVICES = "SERVICES"
    """The physical component represents a set of services."""
    PROCESSES = "PROCESSES"
    """The physical component represents a set of processes."""


@m.stringy_enum
@enum.unique
class PhysicalComponentNature(enum.Enum):
    """The nature of a physical component."""

    UNSET = "UNSET"
    """The physical component nature is not specified."""
    BEHAVIOR = "BEHAVIOR"
    """The physical component nature is behavioral.

    This typically means a piece of software.
    """
    NODE = "NODE"
    """The physical component is a host for behavioral components.

    This typically means a computing resource.
    """


class PhysicalArchitecturePkg(cs.BlockArchitecturePkg):
    packages = m.Containment["PhysicalArchitecturePkg"](
        "ownedPhysicalArchitecturePkgs", (NS, "PhysicalArchitecturePkg")
    )
    architectures = m.Containment["PhysicalArchitecture"](
        "ownedPhysicalArchitectures", (NS, "PhysicalArchitecture")
    )


class PhysicalArchitecture(cs.ComponentArchitecture):
    component_pkg = m.Single["PhysicalComponentPkg"](
        m.Containment(
            "ownedPhysicalComponentPkg", (NS, "PhysicalComponentPkg")
        ),
        enforce=True,
    )
    deployments = m.Containment["cs.AbstractDeploymentLink"](
        "ownedDeployments", (ns.CS, "AbstractDeploymentLink")
    )
    logical_architecture_realizations = m.Containment[
        "LogicalArchitectureRealization"
    ](
        "ownedLogicalArchitectureRealizations",
        (NS, "LogicalArchitectureRealization"),
    )
    realized_logical_architectures = m.Allocation["la.LogicalArchitecture"](
        "ownedLogicalArchitectureRealizations",
        (NS, "LogicalArchitectureRealization"),
        (ns.LA, "LogicalArchitecture"),
        attr="sourceElement",
        backattr="targetElement",
    )

    @property
    def root_function(self) -> PhysicalFunction:
        """Returns the first function in the function_pkg."""
        pkg = self.function_pkg
        assert pkg is not None
        if not pkg.functions:
            raise RuntimeError(f"Package {pkg._short_repr_()} is empty")
        return pkg.functions[0]

    @property
    def root_component(self) -> PhysicalComponent:
        if self.component_pkg is None:
            raise m.BrokenModelError("No root PhysicalComponentPkg found")
        return self.component_pkg.components.by_is_actor(False, single=True)

    @property
    def all_functions(self) -> m.ElementList[PhysicalFunction]:  # type: ignore[override]
        return self._model.search((NS, "PhysicalFunction"), below=self)

    @property
    def all_components(self) -> m.ElementList[PhysicalComponent]:
        return self._model.search((NS, "PhysicalComponent"), below=self)

    @property
    def all_actors(self) -> m.ElementList[PhysicalComponent]:
        return self._model.search((NS, "PhysicalComponent")).by_is_actor(True)

    @property
    def all_function_exchanges(self) -> m.ElementList[fa.FunctionalExchange]:
        return self._model.search((ns.FA, "FunctionalExchange"), below=self)

    @property
    def all_physical_paths(self) -> m.ElementList[cs.PhysicalPath]:
        return self._model.search((ns.CS, "PhysicalPath"), below=self)

    @property
    def all_component_exchanges(self) -> m.ElementList[fa.ComponentExchange]:
        return self._model.search((ns.FA, "ComponentExchange"), below=self)

    @property
    def all_physical_exchanges(self) -> m.ElementList[fa.FunctionalExchange]:
        return self._model.search((ns.FA, "FunctionalExchange"), below=self)

    @property
    def all_physical_links(self) -> m.ElementList[cs.PhysicalLink]:
        return self._model.search((ns.CS, "PhysicalLink"), below=self)

    @property
    def all_functional_chains(self) -> m.ElementList[fa.FunctionalChain]:
        return self._model.search((ns.FA, "FunctionalChain"), below=self)

    diagrams = m.DiagramAccessor(
        "Physical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )

    if not t.TYPE_CHECKING:
        component_package = m.DeprecatedAccessor("component_pkg")


class PhysicalFunction(fa.AbstractFunction):
    """A physical function on the Physical Architecture layer."""

    owner = m.Single["PhysicalComponent"](
        m.Backref((NS, "PhysicalComponent"), "allocated_functions")
    )
    realized_logical_functions = m.Alias["la.LogicalFunction"](
        "realized_functions"
    )
    functions = m.Containment["PhysicalFunction"](
        "ownedPhysicalComponents", (NS, "PhysicalComponent")
    )
    packages = m.Containment["PhysicalFunctionPkg"](
        "ownedPhysicalFunctionPkgs", (NS, "PhysicalFunctionPkg")
    )


class PhysicalFunctionPkg(fa.FunctionPkg):
    """A logical component package."""

    _xmltag = "ownedFunctionPkg"

    functions = m.Containment["PhysicalFunction"](
        "ownedPhysicalFunctions", (NS, "PhysicalFunction")
    )
    packages = m.Containment["PhysicalFunctionPkg"](
        "ownedPhysicalFunctionPkgs", (NS, "PhysicalFunctionPkg")
    )


class PhysicalComponent(
    cs.AbstractPhysicalArtifact,
    cs.Component,
    capellacommon.CapabilityRealizationInvolvedElement,
    cs.DeployableElement,
    cs.DeploymentTarget,
):
    _xmltag = "ownedPhysicalComponents"

    kind = m.EnumPOD("kind", PhysicalComponentKind)
    nature = m.EnumPOD("nature", PhysicalComponentNature)
    deployment_links = m.Containment["cs.AbstractDeploymentLink"](
        "ownedDeploymentLinks", (ns.CS, "AbstractDeploymentLink")
    )

    @property
    def deployed_components(
        self,
    ) -> m.ElementList[PhysicalComponent]:
        return (
            self.parts.map("deployment_links")
            .map("deployed_element")
            .map("type")
        )

    owned_components = m.Containment["PhysicalComponent"](
        "ownedPhysicalComponents", (NS, "PhysicalComponent")
    )
    component_pkgs = m.Containment["PhysicalComponentPkg"](
        "ownedPhysicalComponentPkgs", (NS, "PhysicalComponentPkg")
    )
    deploying_components = m.Backref["PhysicalComponent"](
        (NS, "PhysicalComponent"), "deployed_components"
    )
    allocated_functions = m.Allocation["PhysicalFunction"](
        None, None, (NS, "PhysicalFunction")
    )

    @property
    @deprecated(
        (
            "PhysicalComponent.components is misleadingly named and deprecated;"
            " use 'related_components' instead"
        ),
        category=FutureWarning,
    )
    def components(self) -> m.ElementList[PhysicalComponent]:
        return self.related_components

    @property
    def related_components(self) -> m.ElementList[PhysicalComponent]:
        components = dict.fromkeys(self.deployed_components._elements)
        components.update(dict.fromkeys(self.owned_components._elements))
        return m.ElementList(
            self._model, list(components.keys()), PhysicalComponent
        )

    if not t.TYPE_CHECKING:
        realized_logical_components = m.DeprecatedAccessor(
            "realized_components"
        )


class PhysicalComponentPkg(cs.ComponentPkg, information.AssociationPkg):
    _xmltag = "ownedPhysicalComponentPkg"

    components = m.Containment["PhysicalComponent"](
        "ownedPhysicalComponents", (NS, "PhysicalComponent")
    )
    packages = m.Containment["PhysicalComponentPkg"](
        "ownedPhysicalComponentPkgs", (NS, "PhysicalComponentPkg")
    )
    key_parts = m.Containment["information.KeyPart"](
        "ownedKeyParts", (ns.INFORMATION, "KeyPart")
    )
    deployments = m.Containment["cs.AbstractDeploymentLink"](
        "ownedDeployments", (ns.CS, "AbstractDeploymentLink")
    )


class PhysicalNode(PhysicalComponent):
    pass


class LogicalArchitectureRealization(cs.ArchitectureAllocation):
    pass


class LogicalInterfaceRealization(cs.InterfaceAllocation):
    pass


from .. import la  # noqa: F401
from . import deployment as deployment



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum
import sys
import typing as t
import warnings

import capellambse.model as m

from . import behavior, capellacore, modellingcore
from . import namespaces as ns

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

NS = ns.CAPELLACOMMON


@m.stringy_enum
@enum.unique
class ChangeEventKind(enum.Enum):
    WHEN = "WHEN"


@m.stringy_enum
@enum.unique
class TimeEventKind(enum.Enum):
    AT = "AT"
    """Trigger at a specific time.

    An absolute time trigger is specified with the keyword 'at' followed
    by an expression that evaluates to a time value, such as 'Jan. 1,
    2000, Noon'.
    """
    AFTER = "AFTER"
    """Trigger after a relative time duration has passed.

    A relative time trigger is specified with the keyword 'after'
    followed by an expression that evaluates to a time value, such as
    'after (5 seconds)'.
    """


@m.stringy_enum
@enum.unique
class TransitionKind(enum.Enum):
    INTERNAL = "internal"
    LOCAL = "local"
    EXTERNAL = "external"


class AbstractCapabilityPkg(capellacore.Structure, abstract=True):
    pass


class GenericTrace(capellacore.Trace):
    key_value_pairs = m.Containment["capellacore.KeyValue"](
        "keyValuePairs", (ns.CAPELLACORE, "KeyValue")
    )

    @property
    @deprecated("Synthetic names are deprecated", category=FutureWarning)
    def name(self) -> str:
        myname = type(self).__name__
        if self.target is not None:
            tgname = self.target.name
            tguuid = self.target.uuid
        else:
            tgname = tguuid = "<no target>"
        return f"[{myname}] to {tgname} ({tguuid})"


class TransfoLink(GenericTrace):
    pass


class JustificationLink(GenericTrace):
    pass


class CapabilityRealizationInvolvement(capellacore.Involvement):
    pass


class CapabilityRealizationInvolvedElement(
    capellacore.InvolvedElement, abstract=True
):
    pass


class StateMachine(capellacore.CapellaElement, behavior.AbstractBehavior):
    regions = m.Containment["Region"]("ownedRegions", (NS, "Region"))
    connection_points = m.Containment["Pseudostate"](
        "ownedConnectionPoints", (NS, "Pseudostate")
    )


class Region(capellacore.NamedElement):
    _xmltag = "ownedRegions"

    states = m.Containment["AbstractState"](
        "ownedStates", (NS, "AbstractState")
    )
    transitions = m.Containment["StateTransition"](
        "ownedTransitions", (NS, "StateTransition")
    )
    involved_states = m.Association["AbstractState"](
        (NS, "AbstractState"), "involvedStates"
    )

    __modes = m.Filter["Mode"]("states", (NS, "Mode"))

    @property
    def modes(self) -> m.ElementList[Mode]:
        warnings.warn(
            (
                "Region.modes is deprecated, use states instead"
                " (note that states may only contain either States or Modes)"
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.__modes


class AbstractState(
    capellacore.NamedElement, modellingcore.IState, abstract=True
):
    _xmltag = "ownedStates"

    state_realizations = m.Containment["AbstractStateRealization"](
        "ownedAbstractStateRealizations", (NS, "AbstractStateRealization")
    )
    realized_states = m.Allocation["AbstractState"](
        "ownedAbstractStateRealizations",
        (NS, "AbstractStateRealization"),
        (NS, "AbstractState"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_states = m.Backref["AbstractState"](
        (NS, "AbstractState"), "realized_states"
    )
    incoming_transitions = m.Backref["StateTransition"](
        (NS, "StateTransition"), "target"
    )
    outgoing_transitions = m.Backref["StateTransition"](
        (NS, "StateTransition"), "source"
    )

    if not t.TYPE_CHECKING:

        @property
        def regions(self) -> m.ElementList[Region]:
            myname = AbstractState.__name__
            if type(self) is not AbstractState:
                myname = f"{type(self).__name__}, a subclass of {AbstractState.__name__},"
            warnings.warn(
                (
                    f"{myname} cannot contain regions directly,"
                    f" use the concrete {State.__name__!r} class instead"
                ),
                category=FutureWarning,
                stacklevel=2,
            )
            return m.ElementList(self._model, [], Region)


class State(AbstractState):
    """A situation during which some invariant condition holds.

    A condition of a system or element, as defined by some of its
    properties, which can enable system behaviors and/or structure to
    occur.

    Note: The enabled behavior may include no actions, such as
    associated with a wait state. Also, the condition that defines the
    state may be dependent on one or more previous states.
    """

    regions = m.Containment["Region"]("ownedRegions", (NS, "Region"))
    connection_points = m.Containment["Pseudostate"](
        "ownedConnectionPoints", (NS, "Pseudostate")
    )
    entry = m.Association["behavior.AbstractEvent"](
        (ns.BEHAVIOR, "AbstractEvent"), "entry", legacy_by_type=True
    )
    do_activity = m.Association["behavior.AbstractEvent"](
        (ns.BEHAVIOR, "AbstractEvent"), "doActivity", legacy_by_type=True
    )
    exit = m.Association["behavior.AbstractEvent"](
        (ns.BEHAVIOR, "AbstractEvent"), "exit", legacy_by_type=True
    )
    state_invariant = m.Containment["modellingcore.AbstractConstraint"](
        "stateInvariant", (ns.MODELLINGCORE, "AbstractConstraint")
    )
    functions = m.Backref["fa.AbstractFunction"](
        (ns.FA, "AbstractFunction"), "available_in_states"
    )

    if not t.TYPE_CHECKING:
        entries = m.DeprecatedAccessor("entry")
        exits = m.DeprecatedAccessor("exit")


class Mode(State):
    """Characterizes an expected behavior at a point in time.

    A Mode characterizes an expected behaviour through the set of
    functions or elements available at a point in time.
    """


class FinalState(State):
    """Special state signifying that the enclosing region is completed.

    If the enclosing region is directly contained in a state machine and
    all other regions in the state machine also are completed, then it
    means that the entire state machine is completed.
    """


class StateTransition(capellacore.NamedElement, capellacore.Relationship):
    """A directed relationship between a source and target vertex.

    It may be part of a compound transition, which takes the state
    machine from one state configuration to another, representing the
    complete response of the state machine to an occurrence of an event
    of a particular type.
    """

    _xmltag = "ownedTransitions"

    kind = m.EnumPOD("kind", TransitionKind)
    trigger_description = m.StringPOD("triggerDescription")
    guard = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "guard")
    )
    source = m.Single["AbstractState"](
        m.Association((NS, "AbstractState"), "source")
    )
    target = m.Single["AbstractState"](
        m.Association((NS, "AbstractState"), "target")
    )
    effect = m.Association["behavior.AbstractEvent"](
        (ns.BEHAVIOR, "AbstractEvent"), "effect", legacy_by_type=True
    )
    triggers = m.Association["behavior.AbstractEvent"](
        (ns.BEHAVIOR, "AbstractEvent"), "triggers", legacy_by_type=True
    )
    state_transition_realizations = m.Containment[
        "StateTransitionRealization"
    ]("ownedStateTransitionRealizations", (NS, "StateTransitionRealization"))
    realized_transitions = m.Allocation["StateTransition"](
        "ownedStateTransitionRealizations",
        (NS, "StateTransitionRealization"),
        (NS, "StateTransition"),
        attr="targetElement",
        backattr="sourceElement",
    )

    if not t.TYPE_CHECKING:
        destination = m.DeprecatedAccessor("target")
        effects = m.DeprecatedAccessor("effect")


class Pseudostate(AbstractState, abstract=True):
    pass


class InitialPseudoState(Pseudostate):
    pass


class JoinPseudoState(Pseudostate):
    pass


class ForkPseudoState(Pseudostate):
    pass


class ChoicePseudoState(Pseudostate):
    pass


class TerminatePseudoState(Pseudostate):
    pass


class AbstractStateRealization(capellacore.Allocation):
    pass


class StateTransitionRealization(capellacore.Allocation):
    pass


class ShallowHistoryPseudoState(Pseudostate):
    pass


class DeepHistoryPseudoState(Pseudostate):
    pass


class EntryPointPseudoState(Pseudostate):
    pass


class ExitPointPseudoState(Pseudostate):
    pass


class StateEventRealization(capellacore.Allocation):
    pass


class StateEvent(
    capellacore.NamedElement, behavior.AbstractEvent, abstract=True
):
    expression = m.Association["capellacore.Constraint"](
        (ns.CAPELLACORE, "Constraint"), "expression"
    )
    state_event_realizations = m.Containment["StateEventRealization"](
        "ownedStateEventRealizations", (NS, "StateEventRealization")
    )
    realized_events = m.Allocation["StateEvent"](
        "ownedStateEventRealizations",
        (NS, "StateEventRealization"),
        (NS, "StateEvent"),
        attr="targetElement",
        backattr="sourceElement",
    )


class ChangeEvent(StateEvent):
    kind = m.EnumPOD("kind", ChangeEventKind)


class TimeEvent(StateEvent):
    kind = m.EnumPOD("kind", TimeEventKind)


if not t.TYPE_CHECKING:

    def __getattr__(name):
        if name == "AbstractStateMode":
            warnings.warn(
                "AbstractStateMode has been renamed to AbstractState",
                DeprecationWarning,
                stacklevel=2,
            )
            return AbstractState

        raise AttributeError(name)


from . import fa  # noqa: F401



#!/usr/bin/env python
# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import collections
import dataclasses
import enum
import logging
import operator
import re
import sys
import textwrap
import typing as t

import click
from lxml import etree

from capellambse import model
from capellambse.helpers import qtype_of

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

LOGGER = logging.getLogger(__name__)

NS_ECORE = "http://www.eclipse.org/emf/2002/Ecore"
EPackage = etree.QName(NS_ECORE, "EPackage")
EEnum = etree.QName(NS_ECORE, "EEnum")
EClass = etree.QName(NS_ECORE, "EClass")
EReference = etree.QName(NS_ECORE, "EReference")
EAttribute = etree.QName(NS_ECORE, "EAttribute")

DOCUMENTATION_SOURCE = "http://www.polarsys.org/kitalpha/ecore/documentation"
XPATH_DESCRIPTION = (
    "./eAnnotations[@source='{}']/details[@key='{}']/@value"
).format(DOCUMENTATION_SOURCE, "{}")
BAD_DESCRIPTIONS = ("n/a", "none")


@click.command()
@click.argument("ecore_file", type=click.File("r"))
@click.argument("python_file", type=click.File("w"))
@click.option(
    "--subpackage",
    nargs=1,
    default=None,
    help=(
        "Write a module for the given subpackage."
        " Use '.' for the top-level package (equivalent to __init__.py)."
    ),
)
@click.option(
    "--docstrings/--no-docstrings",
    is_flag=True,
    default=True,
    help="Add docstrings with documentation extracted from the XML.",
)
@click.option(
    "--namespaces-module",
    is_flag=True,
    default=False,
    help=(
        "Assume a module named 'namespaces' containing all Namespace"
        " instances. If not passed, each imported module is assumed to have a"
        " top-level 'NS' attribute containing the respective Namespace"
        " instance."
    ),
)
def main(
    *,
    ecore_file: t.IO[str],
    python_file: t.IO[str],
    subpackage: str | None,
    docstrings: bool,
    namespaces_module: bool,
) -> None:
    """Generate a Python module from an ecore definition.

    Note that the code is generated on a best-effort basis, with no hard
    guarantees about syntactical correctness. Also, due to the script
    only handling a single file, import locations of other modules will
    likely need to be adjusted as well.

    Only basic code formatting is applied, mainly for syntactical
    reasons. It's recommended to run an auto-formatter over the produced
    output file.
    """
    logging.basicConfig(level="INFO")
    tree = etree.parse(ecore_file)
    root = tree.getroot()
    if qtype_of(root) != EPackage:
        raise click.UsageError(
            "Provided ecore file does not contain an EPackage as root element"
        )

    module_doc = find_description(root)

    imports: set[str] = set()
    classes = collections.deque[ClassDef]()
    enums: list[EnumDef] = []
    location_path: list[str] = []

    if subpackage is not None and subpackage != ".":
        for i in subpackage.split("."):
            location_path.append(i)
            pkgs = root.xpath(f"./eSubpackages[@name={i!r}]")
            if len(pkgs) > 1:
                raise SystemExit(
                    "Error: Multiple matches for subpackage"
                    f" {i!r} in {root.get('name')}"
                )
            if len(pkgs) < 1:
                raise SystemExit(f"Error: Subpackage not found: {subpackage}")
            (root,) = pkgs
    location = ".".join(location_path)

    for classifier in root.iterchildren("eClassifiers"):
        qtype = qtype_of(classifier)
        if qtype == EEnum:
            enums.append(parse_enum(classifier))
        elif qtype == EClass:
            cls, new_imports = parse_class(location, classifier)
            classes.append(cls)
            imports |= new_imports

    with python_file as f:
        if docstrings and module_doc:
            writedoc(f, module_doc, 0)

        f.write("from __future__ import annotations\n")
        if enums:
            f.write("\nimport enum\n")
        f.write("\nimport capellambse.model as m\n")

        if imports:
            if namespaces_module:
                imports.add("namespaces as ns")
            if subpackage is None:
                isource = "."
            elif subpackage == ".":
                isource = ".."
            else:
                isource = "." * (1 + subpackage.count("."))
            f.write(f"\nfrom {isource} import {', '.join(sorted(imports))}\n")

        # fmt: off
        f.write(
            (
                "\n"
                "NS = m.Namespace(\n"
                "    {nsURI!r},\n"
                "    {nsPrefix!r},\n"
                ")\n"
            ).format(**root.attrib)
        )
        # fmt: on

        write_enums(f, enums, docstrings)
        write_classes(f, location, classes, docstrings, namespaces_module)


def write_enums(
    f: t.IO[str],
    enums: list[EnumDef],
    docstrings: bool,
) -> None:
    for enumdef in sorted(enums, key=operator.attrgetter("name")):
        f.write("\n\n@m.stringy_enum\n@enum.unique\n")
        f.write(f"class {enumdef.name}(enum.Enum):\n")
        if docstrings and enumdef.docstring:
            writedoc(f, enumdef.docstring, 1)
            if enumdef.literals:
                f.write("\n")

        for lit in enumdef.literals:
            f.write(f"    {lit.name} = {lit.value!r}\n")
        if (not docstrings or not enumdef.docstring) and not enumdef.literals:
            f.write("    pass\n")


def write_classes(
    f: t.IO[str],
    location: str,
    classes: collections.deque[ClassDef],
    docstrings: bool,
    nsmodule: bool,
) -> None:
    written_classes: set[str] = set()
    while classes:
        cls = classes.popleft()

        for base in cls.bases:
            if "." in base or base in written_classes:
                continue
            classes.appendleft(cls)
            try:
                index, basedef = next(
                    (i, c) for i, c in enumerate(classes) if c.name == base
                )
            except StopIteration:
                if location:
                    isource = "." * (1 + location.count("."))
                    f.write(f"\n\nfrom {isource} import {base}")
                else:
                    raise ValueError(
                        f"Reference to undefined base {base}"
                    ) from None
                written_classes.add(base)
            else:
                del classes[index]
                classes.appendleft(basedef)
                break

        else:
            write_class(f, cls, docstrings=docstrings, nsmodule=nsmodule)
            written_classes.add(cls.name)


def write_class(
    f: t.IO[str],
    cls: ClassDef,
    *,
    docstrings: bool,
    nsmodule: bool,
) -> None:
    f.write(f"\n\nclass {cls.name}(")
    f.write(", ".join(cls.bases) or "m.ModelElement")
    if cls.abstract:
        f.write(", abstract=True")
    f.write("):\n")

    if docstrings and cls.docstring:
        writedoc(f, cls.docstring, 1)
        if cls.members:
            f.write("\n")

    for member in cls.members:
        if isinstance(member, EnumPODMember):
            f.write(
                f"    {member.name} = m.EnumPOD"
                f"({member.xmlname!r}, {member.pod_type})\n"
            )
        elif isinstance(member, PODMember):
            f.write(
                f"    {member.name} = m.{member.pod_type}POD"
                f"({member.xmlname!r})\n"
            )
        elif isinstance(member, RelationshipMember):
            write_relationship(f, member, nsmodule=nsmodule)
        else:
            raise AssertionError(
                f"Unhandled member type {type(member).__name__}"
            )
        if docstrings and member.docstring:
            writedoc(f, member.docstring, 1)

    if not cls.members and not (docstrings and cls.docstring):
        f.write("    pass\n")


def write_relationship(
    f: t.IO[str],
    member: RelationshipMember,
    *,
    nsmodule: bool,
) -> None:
    f.write(f"    {member.name} = ")

    member_type = member.type.value
    if member.single:
        f.write(f"m.Single[{member.cls!r}](m.{member_type}(")
    else:
        f.write(f"m.{member_type}[{member.cls!r}](")

    clsname = clsname2tuplestring(member.cls, nsmodule)
    match member.type:
        case RelationshipType.Containment:
            f.write(f"{member.xmlname!r}, {clsname}")
        case RelationshipType.Association:
            f.write(f"{clsname}, {member.xmlname!r}")
        case _:
            assert_never(f"Unhandled relationship type: {member.type.name}")

    if member.single:
        f.write(")")
    f.write(")\n")


def writedoc(f, docstring: str, indent: int) -> None:
    wrapped = f'"""{docstring}\n"""\n'
    f.write(textwrap.indent(wrapped, "    " * indent))


def find_description(element: etree._Element) -> str:
    description = [
        i.strip().replace("\r\n", "\n")
        for i in element.xpath(XPATH_DESCRIPTION.format("description"))
        if not i.lower().startswith(BAD_DESCRIPTIONS)
    ]
    assert all(isinstance(i, str) for i in description)

    usage_guideline = [
        i.strip().replace("\r\n", "\n")
        for i in element.xpath(XPATH_DESCRIPTION.format("usage guideline"))
        if not i.strip().lower().startswith(BAD_DESCRIPTIONS)
    ]
    if usage_guideline:
        description.append("Usage guideline\n---------------\n")
        description.extend(usage_guideline)
    return "\n\n".join(description)


@dataclasses.dataclass(frozen=True)
class EnumDef:
    name: str
    docstring: str
    literals: tuple[EnumLiteral, ...]


@dataclasses.dataclass(frozen=True)
class EnumLiteral:
    name: str
    docstring: str
    value: str


def parse_enum(element: etree._Element) -> EnumDef:
    literal_elems = sorted(
        element.iterchildren("eLiterals"),
        key=lambda i: int(i.get("value", "0")),
    )
    return EnumDef(
        name=element.attrib["name"],
        docstring=find_description(element),
        literals=tuple(
            EnumLiteral(
                name=screaming_snake(lit.attrib["name"]),
                docstring=find_description(lit),
                value=lit.attrib["name"],
            )
            for lit in literal_elems
        ),
    )


@dataclasses.dataclass(frozen=True)
class ClassDef:
    name: str
    docstring: str
    bases: tuple[str, ...]
    members: tuple[ClassMember, ...]
    abstract: bool


@dataclasses.dataclass(frozen=True)
class ClassMember:
    name: str
    docstring: str


@dataclasses.dataclass(frozen=True)
class PODMember(ClassMember):
    pod_type: str
    xmlname: str


@dataclasses.dataclass(frozen=True)
class EnumPODMember(PODMember):
    pass


@dataclasses.dataclass(frozen=True)
class RelationshipMember(ClassMember):
    type: RelationshipType
    cls: str
    single: bool
    xmlname: str


class RelationshipType(enum.Enum):
    Containment = model.Containment.__name__
    Association = model.Association.__name__


def parse_class(
    location: str, element: etree._Element
) -> tuple[ClassDef, set[str]]:
    imports: set[str] = set()
    bases: list[str] = []
    for base in element.get("eSuperTypes", "").split():
        name = clspath2dotted(base, start=location)
        if "." in name:
            imports.add(name.split(".", 1)[0])
        bases.append(name)

    members: list[ClassMember] = []
    clsname = element.attrib["name"]
    for member in element.iterchildren("eStructuralFeatures"):
        memberdef = parse_class_member(location, clsname, member)
        if memberdef is not None:
            members.append(memberdef)

    classdef = ClassDef(
        name=clsname,
        docstring=find_description(element),
        bases=tuple(bases),
        members=tuple(members),
        abstract=element.get("abstract", "false") == "true",
    )
    return classdef, imports


def parse_class_member(
    location: str,
    clsname: str,
    member: etree._Element,
) -> ClassMember | None:
    membname = member.attrib["name"]
    pyname = camel2snake(member.attrib["name"])
    qtype = qtype_of(member)
    if qtype == EReference:
        if member.get("derived", "false") == "true":
            LOGGER.warning(
                "Ignoring derived class member %s.%s", clsname, membname
            )
            return None

        clsname = clspath2dotted(
            member.attrib["eType"].rsplit(" ", 1)[-1], start=location
        )
        single = (
            member.get("upperBound", "0") == "1"
            or member.get("lowerBound", "0") == "1"
        )

        if member.get("containment", "false") == "true":
            return RelationshipMember(
                type=RelationshipType.Containment,
                name=camel2snake(membname).removeprefix("owned_"),
                docstring=find_description(member),
                xmlname=membname,
                cls=clsname,
                single=single,
            )

        return RelationshipMember(
            type=RelationshipType.Association,
            name=camel2snake(membname),
            docstring=find_description(member),
            xmlname=membname,
            cls=clsname,
            single=single,
        )

    if qtype == EAttribute:
        attrtype = clspath2dotted(
            member.attrib["eType"].rsplit(" ", 1)[-1], start=location
        )
        cls = PODMember
        if attrtype == "ecore.EBoolean":
            pod_type = "Bool"
            if pyname != "value":
                pyname = "is_" + pyname.removeprefix("is_")
        elif attrtype in {"ecore.EString", "ecore.EChar"}:
            pod_type = "String"
        elif attrtype in {
            "ecore.EByte",
            "ecore.EInt",
            "ecore.EBigInteger",
            "ecore.ELong",
        }:
            pod_type = "Int"
        elif attrtype in {"ecore.EDouble", "ecore.EFloat"}:
            pod_type = "Float"
        elif attrtype == "ecore.EDate":
            pod_type = "Datetime"
        else:
            cls = EnumPODMember
            pod_type = attrtype
        return cls(
            name=pyname,
            docstring=find_description(member),
            pod_type=pod_type,
            xmlname=member.attrib["name"],
        )

    LOGGER.warning("Unknown structural feature type: %s", qtype)
    return None


def screaming_snake(name: str) -> str:
    return re.sub(
        r"[a-z]+",
        lambda m: m.group(0).upper() + "_" * (m.span(0)[1] != len(name)),
        name,
    )


def camel2snake(name: str) -> str:
    return re.sub(
        "[A-Z]",
        lambda m: "_" * (m.start(0) > 0) + m.group(0).lower(),
        name,
    )


def clspath2dotted(path: str, *, start: str) -> str:
    module, _, name = path.rpartition("#//")
    name = name.replace("/", ".")
    if module:
        return f"{fname2modname(module)}.{name}"

    if name.startswith(f"{start}."):
        name = name.removeprefix(start).lstrip(".")
    return name


def fname2modname(fname: str) -> str:
    return fname.rpartition("/")[2].removesuffix(".ecore").lower()


def clsname2tuplestring(clsname: str, nsmodule: bool) -> str:
    if "." in clsname:
        module, clsname = clsname.rsplit(".", 1)
        if nsmodule:
            return f"(ns.{module.upper()}, {clsname!r})"
        return f"({module}.NS, {clsname!r})"
    return f"(NS, {clsname!r})"


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Common namespace definitions for the Capella metamodel."""

import capellambse.model as m

MODELLINGCORE = m._obj.NS
METADATA = m._obj.NS_METADATA

ACTIVITY = m.Namespace(
    "http://www.polarsys.org/capella/common/activity/{VERSION}",
    "org.polarsys.capella.common.data.activity",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
BEHAVIOR = m.Namespace(
    "http://www.polarsys.org/capella/common/behavior/{VERSION}",
    "org.polarsys.capella.common.data.behavior",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
CAPELLACOMMON = m.Namespace(
    "http://www.polarsys.org/capella/core/common/{VERSION}",
    "org.polarsys.capella.core.data.capellacommon",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
CAPELLACORE = m.Namespace(
    "http://www.polarsys.org/capella/core/core/{VERSION}",
    "org.polarsys.capella.core.data.capellacore",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
CAPELLAMODELLER = m.Namespace(
    "http://www.polarsys.org/capella/core/modeller/{VERSION}",
    "org.polarsys.capella.core.data.capellamodeller",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
CS = m.Namespace(
    "http://www.polarsys.org/capella/core/cs/{VERSION}",
    "org.polarsys.capella.core.data.cs",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
EPBS = m.Namespace(
    "http://www.polarsys.org/capella/core/epbs/{VERSION}",
    "org.polarsys.capella.core.data.epbs",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
FA = m.Namespace(
    "http://www.polarsys.org/capella/core/fa/{VERSION}",
    "org.polarsys.capella.core.data.fa",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
INFORMATION = m.Namespace(
    "http://www.polarsys.org/capella/core/information/{VERSION}",
    "org.polarsys.capella.core.data.information",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
INFORMATION_COMMUNICATION = m.Namespace(
    "http://www.polarsys.org/capella/core/information/communication/{VERSION}",
    "org.polarsys.capella.core.data.information.communication",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
INFORMATION_DATATYPE = m.Namespace(
    "http://www.polarsys.org/capella/core/information/datatype/{VERSION}",
    "org.polarsys.capella.core.data.information.datatype",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
INFORMATION_DATAVALUE = m.Namespace(
    "http://www.polarsys.org/capella/core/information/datavalue/{VERSION}",
    "org.polarsys.capella.core.data.information.datavalue",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
INTERACTION = m.Namespace(
    "http://www.polarsys.org/capella/core/interaction/{VERSION}",
    "org.polarsys.capella.core.data.interaction",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
LA = m.Namespace(
    "http://www.polarsys.org/capella/core/la/{VERSION}",
    "org.polarsys.capella.core.data.la",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
LIBRARIES = m.Namespace(
    "http://www.polarsys.org/capella/common/libraries/{VERSION}",
    "libraries",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
OA = m.Namespace(
    "http://www.polarsys.org/capella/core/oa/{VERSION}",
    "org.polarsys.capella.core.data.oa",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
PA = m.Namespace(
    "http://www.polarsys.org/capella/core/pa/{VERSION}",
    "org.polarsys.capella.core.data.pa",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
PA_DEPLOYMENT = m.Namespace(
    "http://www.polarsys.org/capella/core/pa/deployment/{VERSION}",
    "org.polarsys.capella.core.data.pa.deployment",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
RE = m.Namespace(
    "http://www.polarsys.org/capella/common/re/{VERSION}",
    "re",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
SA = m.Namespace(
    "http://www.polarsys.org/capella/core/ctx/{VERSION}",
    "org.polarsys.capella.core.data.ctx",
    m.CORE_VIEWPOINT,
    "7.0.0",
)
SHARED_MODEL = m.Namespace(
    "http://www.polarsys.org/capella/core/sharedmodel/{VERSION}",
    "org.polarsys.capella.core.data.sharedmodel",
    m.CORE_VIEWPOINT,
    "7.0.0",
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

__all__ = ["MelodyModel", "ModelInfo"]

import collections.abc as cabc
import dataclasses
import itertools
import logging
import os
import typing as t
import warnings

from lxml import etree

import capellambse
import capellambse.helpers
from capellambse import _diagram_cache, aird, filehandler, loader

from . import _descriptors, _obj, diagram

# `pathlib` is referenced by the `dataclasses` auto-generated init.
# Sphinx-autodoc-typehints needs this import here to resolve the forward ref.
import pathlib  # isort: skip  # noqa: F401

LOGGER = logging.getLogger(__name__)


class MelodyModel:
    """Provides high-level access to a model.

    This class builds upon the lower-level
    :class:`~capellambse.loader.core.MelodyLoader` to provide an
    abstract, high-level interface for easy access to various model
    aspects.
    """

    @property
    def project(self) -> capellambse.metamodel.capellamodeller.Project:
        import capellambse.metamodel as mm  # noqa: PLC0415

        cls_project = self.resolve_class((mm.capellamodeller.NS, "Project"))
        if __debug__:
            cls_library = self.resolve_class(
                (mm.capellamodeller.NS, "Library")
            )
            assert issubclass(cls_library, cls_project)

        roots: list[etree._Element] = []
        for fname, frag in self._loader.trees.items():
            if fname.parts[0] != "\x00":
                continue
            elem = frag.root
            if elem.tag == "{http://www.omg.org/XMI}XMI":
                try:
                    elem = next(iter(elem))
                except StopIteration:
                    continue
            qname = capellambse.helpers.qtype_of(elem)
            if qname is None:
                continue
            try:
                elemcls = self.resolve_class(qname)
            except (_obj.UnknownNamespaceError, _obj.MissingClassError):
                continue
            if not issubclass(elemcls, cls_project):
                continue
            roots.append(elem)

        if len(roots) < 1:
            raise RuntimeError("No root project or library found")
        if len(roots) > 1:
            raise RuntimeError(f"Found {len(roots)} Project/Library objects")
        obj = _obj.wrap_xml(self, roots[0])
        assert isinstance(obj, mm.capellamodeller.Project)
        return obj

    @property
    def oa(self) -> capellambse.metamodel.oa.OperationalAnalysis:
        return self.project.model_root.oa

    @property
    def sa(self) -> capellambse.metamodel.sa.SystemAnalysis:
        return self.project.model_root.sa

    @property
    def la(self) -> capellambse.metamodel.la.LogicalArchitecture:
        return self.project.model_root.la

    @property
    def pa(self) -> capellambse.metamodel.pa.PhysicalArchitecture:
        return self.project.model_root.pa

    @property
    def epbs(self) -> capellambse.metamodel.epbs.EPBSArchitecture:
        return self.project.model_root.epbs

    def enumeration_property_types(
        self,
    ) -> _obj.ElementList[
        capellambse.metamodel.capellacore.EnumerationPropertyType
    ]:
        return self.project.model_root.enumeration_property_types

    def property_value_packages(
        self,
    ) -> _obj.ElementList[capellambse.metamodel.capellacore.PropertyValuePkg]:
        return self.project.model_root.property_value_packages

    @property
    def uuid(self) -> str:
        """The unique ID of the model's root element."""
        return self.project.uuid

    @property
    def name(self) -> str:
        """The name of this model."""
        return self.project.name

    @name.setter
    def name(self, value: str) -> None:
        self.project.name = value

    diagrams = diagram.DiagramAccessor(
        None, cacheattr="_MelodyModel__diagram_cache"
    )

    diagram_cache: filehandler.FileHandler | None

    _diagram_cache_index: (
        dict[str, capellambse._diagram_cache.IndexEntry] | None
    )
    """An index describing the diagrams in the cache.

    If the attribute does not exist, the index has not been loaded from
    the file handler yet. If the attribute is None, it means that the
    index file could not be found and all diagram files need to tried.
    """
    __diagram_cache: t.Any
    """Stores proxy instances for Diagram.

    Not to be confused with the public "diagram_cache", which is where
    the pre-rendered content for those instances is fetched from.
    """

    def __init__(
        self,
        path: str | os.PathLike,
        *,
        diagram_cache: (
            str
            | os.PathLike
            | filehandler.FileHandler
            | dict[str, t.Any]
            | None
        ) = None,
        fallback_render_aird: bool = False,
        **kwargs: t.Any,
    ) -> None:
        """Load a project.

        For complete information on which exact ``kwargs`` are
        supported, consult the documentation of the used file handler.
        Refer to the "See Also" section for a collection of links.

        Below are some common parameter names and their usual meanings.
        Not all file handlers support all parameters.

        .. note::
            Passing in arguments that are not accepted by the selected
            file handler will result in an exception being raised.
            Similarly, leaving out arguments that are required by the
            file handler will also result in an exception.

        Parameters
        ----------
        path
            Path or URL to the project. The following formats are
            accepted:

            * A path to a local ``.aird`` file.
            * A path to a local directory (requires ``entrypoint``).
            * An SCP-style short URL, which will be treated as referring
              to a Git repository.

              Example: ``git@github.com:dbinfrago/py-capellambse.git``
            * A remote URL, with a protocol or prefix that indicates
              which file handler to invoke (requires ``entrypoint``).

              Some examples:

              * ``git://git.example.com/model/coffeemaker.git``
              * ``git+https://git.example.com/model/coffeemaker.git``
              * ``git+ssh://git@git.example.com/model/coffeemaker.git``

              .. note:: Depending on the exact file handler, saving back
                 to a remote location might fail with ``update_cache``
                 set to ``False``. See :meth:`save` for more details.
        entrypoint: str
            Entrypoint from path to the main ``.aird`` file.
        revision: str
            The revision to use, if loading a model from a version
            control system like git. Defaults to the current HEAD. If
            the used VCS does not have a notion of "current HEAD", this
            argument is mandatory.
        disable_cache: bool
            Disable local caching of remote content.
        update_cache: bool
            Update the local cache. Defaults to ``True``, but can be
            disabled to reuse the last cached state.
        identity_file: str | pathlib.Path
            The identity file (private key) to use when connecting via
            SSH.
        known_hosts_file: str | pathlib.Path
            The ``known_hosts`` file to pass to SSH for verifying the
            server's host key.
        username: str
            The username to log in as remotely.
        password: str
            The password to use for logging in. Will be ignored when
            ``identity_file`` is passed as well.
        diagram_cache
            An optional place where to find pre-rendered, cached
            diagrams. When a diagram is found in this cache, it will be
            loaded from there instead of being rendered on access. Note
            that diagrams will only be loaded from there, but not be put
            back, i.e. to use it effectively, the cache has to be
            pre-populated.

            This argument accepts the following values:

            - ``None``, in which case all diagrams will be rendered
              internally the first time they are used.
            - A path to a local directory.
            - A URL just like for the ``path`` argument.
            - A dictionary with the arguments to a
              :py:class:`~capellambse.filehandler.abc.FileHandler`. The
              dict's ``path`` key will be analyzed to determine the
              correct FileHandler class.
            - An instance of
              :class:`~capellambse.filehandler.abc.FileHandler`, which
              will be used directly.

            .. warning:: When using the diagram cache, always make sure
               that the cached diagrams actually match the model version
               that is being used. There is no way to check this
               automatically.

            The file names looked up in the cache built in the format
            ``uuid.ext``, where ``uuid`` is the UUID of the diagram (as
            reported by ``diag_obj.uuid``) and ``ext`` is the render
            format. Example:

            - Diagram ID: ``_7FWu4KrxEeqOgqWuHJrXFA``
            - Render call: ``diag_obj.as_svg`` or ``diag_obj.render("svg")``
            - Cache file name: ``_7FWu4KrxEeqOgqWuHJrXFA.svg``

            *This argument is **not** passed to the file handler.*
        fallback_render_aird: bool
            If set to True, always fall back to rendering diagrams
            internally if the configured ``diagram_cache`` is not
            available.

            By default, the internal renderer is entirely disabled for
            AIRD diagrams, and only used during cache misses
            (``FileNotFoundError`` from the underlying file handler) for
            all other types of diagrams.
        **kwargs
            Additional arguments are passed on to the underlying
            :class:`~capellambse.loader.core.MelodyLoader`, which in
            turn passes it on to the primary resource's FileHandler.

        See Also
        --------
        capellambse.filehandler.FileHandler :
            Abstract super class for file handlers. Contains information
            needed for implementing custom handlers.
        capellambse.filehandler.local.LocalFileHandler :
            The file handler responsible for local files and
            directories.
        capellambse.filehandler.git.GitFileHandler :
            The file handler implementing the ``git://`` protocol.
        capellambse.filehandler.http.HTTPFileHandler :
            A simple ``http(s)://`` file handler.
        """
        capellambse.load_model_extensions()

        self._loader = loader.MelodyLoader(path, **kwargs)
        self.__viewpoints = dict(self._loader.referenced_viewpoints())
        self._fallback_render_aird = fallback_render_aird

        if diagram_cache:
            if diagram_cache == path:
                self.diagram_cache = self._loader.filehandler
            elif isinstance(diagram_cache, filehandler.FileHandler):
                self.diagram_cache = diagram_cache
            elif isinstance(diagram_cache, cabc.Mapping):
                self.diagram_cache = filehandler.get_filehandler(
                    **diagram_cache
                )
            else:
                self.diagram_cache = filehandler.get_filehandler(diagram_cache)
        else:
            self.diagram_cache = None

    @property
    def resources(self) -> dict[str, filehandler.FileHandler]:
        return self._loader.resources

    def save(self, **kw: t.Any) -> None:
        """Save the model back to where it was loaded from.

        Parameters
        ----------
        kw
            Additional keyword arguments accepted by the file handler in
            use. Please see the respective documentation for more info.

        See Also
        --------
        capellambse.filehandler.localfilehandler.LocalFileHandler.write_transaction :
            Accepted ``**kw`` when using local directories
        capellambse.filehandler.git.GitFileHandler.write_transaction :
            Accepted ``**kw`` when using ``git://`` and similar URLs

        Notes
        -----
        With a file handler that contacts a remote location (such as the
        :class:`~capellambse.filehandler.git.GitFileHandler`
        with non-local repositories), saving might fail if the local
        state has gone out of sync with the remote state. To avoid this,
        always leave the ``update_cache`` parameter at its default value
        of ``True`` if you intend to save changes.
        """
        self._loader.save(**kw)

    def search(
        self,
        *clsnames: str | type[_obj.ModelObject] | _obj.UnresolvedClassName,
        below: _obj.ModelElement | None = None,
    ) -> _obj.ElementList:
        """Search for all elements with any of the given types.

        Parameters
        ----------
        clsnames
            The classes to search for. Instances of subclasses of any of
            the given types will also be returned. If no class is
            specified, all model elements are returned.

            Each element can be:

            - A string containing the name of a class. All namespaces
              belonging to an activated viewpoint are searched for
              classes with that name.
            - A "namespace.alias:ClassName" string.
            - A (Namespace, 'ClassName') tuple.
            - A ('namespace.alias', 'ClassName') tuple.
            - A Class object. It must have the ``__capella_namespace__``
              attribute set to its namespace, which is automatically
              done for subclasses of
              :class:`~capellambse.model.ModelElement`.

            .. note::

               This method treats the
               :class:`~capellambse.model.ModelElement`
               class as the superclass of every concrete model element
               class. This means that any search query including it will
               return all model elements regardless of their type.
        below
            A model element to constrain the search. If given, only
            those elements will be returned that are (immediate or
            nested) children of this element. This option takes into
            account model fragmentation, but it does not treat link
            elements specially.

        Notes
        -----
        For performance reasons, this method only takes into account
        semantic fragments and diagram descriptors.

        Examples
        --------
        The following calls are functionally identical, and will all
        return a list of every Logical Component in the model:

        >>> model.search("LogicalComponent")
        >>> model.search("org.polarsys.capella.core.data.la:LogicalComponent")
        >>> model.search( (capellambse.metamodel.la.NS, "LogicalComponent") )
        >>> model.search( ("org.polarsys.capella.core.data.la", "LogicalComponent") )
        """
        classes: set[type[_obj.ModelObject]] = set()
        for clsname in clsnames:
            if isinstance(clsname, type):
                if not hasattr(clsname, "__capella_namespace__"):
                    raise TypeError(
                        f"Class does not belong to a namespace: {clsname!r}"
                    )
                resolved = self.resolve_class(
                    (clsname.__capella_namespace__, clsname.__name__)
                )
            elif clsname == "GenericElement":
                warnings.warn(
                    "GenericElement has been renamed to ModelElement",
                    DeprecationWarning,
                    stacklevel=2,
                )
                resolved = _obj.ModelElement
            elif clsname == "ModelObject":
                resolved = _obj.ModelElement
            else:
                resolved = self.resolve_class(clsname)
            if resolved is _obj.ModelElement:
                classes.clear()
                break
            classes.add(resolved)

        trees = [
            t
            for t in self._loader.trees.values()
            if t.fragment_type is loader.FragmentType.SEMANTIC
        ]
        matches: cabc.Iterable[etree._Element]
        if not classes:
            matches = itertools.chain.from_iterable(
                tree.iterall() for tree in trees
            )
        else:
            matches = []
            for tree in trees:
                for qtype in tree.iter_qtypes():
                    try:
                        cls = self.resolve_class(qtype)
                    except (
                        _obj.UnknownNamespaceError,
                        _obj.MissingClassError,
                    ):
                        continue
                    if any(issubclass(cls, i) for i in classes):
                        matches.extend(tree.iter_qtype(qtype))

        if not classes or diagram.Diagram in classes:
            matches = itertools.chain(
                matches,
                aird.enumerate_descriptors(self._loader),
            )

        if below is not None:
            matches = (
                i
                for i in matches
                if below._element in self._loader.iterancestors(i)
            )
        seen: set[int] = set()
        elements: list[etree._Element] = []
        for elem in matches:
            if id(elem) not in seen:
                elements.append(elem)
                seen.add(id(elem))
            else:
                LOGGER.warning("Found element twice (bad caches?): %r", elem)
        return _obj.ElementList(
            self,
            elements,
            _obj.ModelElement,
            legacy_by_type=len(clsnames) != 1,
        )

    def by_uuid(self, uuid: str) -> t.Any:
        """Search the entire model for an element with the given UUID."""
        return _obj.wrap_xml(self, self._loader[uuid])

    def find_references(
        self, target: _obj.ModelObject | str, /
    ) -> cabc.Iterator[tuple[t.Any, str, int | None]]:
        """Search the model for references to the given object.

        Parameters
        ----------
        target
            The target object to search for.

        Yields
        ------
        tuple[ModelObject, str, int | None]
            A 3-tuple containing the referencing model object, the
            attribute on that object, and an optional index. If the
            attribute contains a list of objects, the index shows the
            index into the list that was found. Otherwise the index is
            None.
        """
        if isinstance(target, str):
            if not capellambse.helpers.is_uuid_string(target):
                raise ValueError(f"Malformed UUID: {target!r}")
            uuid: str = target
            target = self.by_uuid(target)

        else:
            if target._model is not self:
                raise ValueError(
                    "Cannot find references to objects from different models"
                )
            uuid = getattr(target, "uuid", "")
            if not capellambse.helpers.is_uuid_string(uuid):
                raise ValueError(f"Malformed or missing UUID for {target!r}")

        for elem in self._loader.xpath(
            f"//*[@*[contains(., '#{uuid}')] | */@*[contains(., '#{uuid}')]]",
            roots=[
                i.root
                for i in self._loader.trees.values()
                if i.fragment_type != loader.FragmentType.VISUAL
            ],
        ):
            obj = _obj.wrap_xml(self, elem)
            for attr in _reference_attributes(type(obj)):
                if attr.startswith("_"):
                    continue

                try:
                    value = getattr(obj, attr)
                except Exception:
                    continue

                if isinstance(value, _obj.ModelElement) and value == target:
                    yield (obj, attr, None)
                elif isinstance(value, _obj.ElementList):
                    try:
                        idx = value.index(target)
                    except ValueError:
                        continue
                    yield (obj, attr, idx)

    def update_diagram_cache(
        self,
        capella_cli: str,
        image_format: t.Literal["bmp", "gif", "jpg", "png", "svg"] = "svg",
        *,
        create_index: bool = False,
        force: t.Literal["docker", "exe"] | None = None,
        background: bool = True,
    ) -> None:
        r"""Update the diagram cache if one has been specified.

        If a ``diagram_cache`` has been specified while loading a
        Capella model it will be updated when this function is called.

        The diagram cache will be populated by executing the Capella
        function "Export representations as images" which is normally
        accessible via the context menu of an ``.aird`` node in
        Capella's project explorer. The export of diagrams happens with
        the help of Capella's command line interface (CLI).

        The CLI of Capella must be specified by the caller. It is
        possible to work with a local installation or a Docker image of
        an individual Capella bundle.

        At the moment it is supported to run a Docker image using the
        container system Docker and the ``docker`` executable must be in
        the ``PATH`` environment variable.

        Parameters
        ----------
        capella_cli
            The Capella CLI to use when exporting diagrams from the
            given Capella model. The provided string can come with a
            ``"{VERSION}"`` placeholder. If specified, this placeholder
            will be replaced by the x.y.z formatted version of Capella
            that has been used when the given Capella model was last
            saved. After consideration of the optional placeholder this
            function will first check if the value of ``capella_cli``
            points to a local Capella executable (that can be an
            absolute path or an executable/ symbolic link that has been
            made available via the environment variable ``PATH``). If no
            executable can be found it is expected that ``capella_cli``
            represents a Docker image name for an image that behaves
            like the Capella CLI. For the case of passing a Docker image
            name through ``capella_cli`` this means it is assumed that
            something like the following

            .. code-block:: bash

                docker run --rm -it <capella_cli> -nosplash \
                    -consolelog -app APP -appid APPID

            will work.
            The parameter ``force`` can be set to change the described
            behaviour and force the function to treat the
            ``capella_cli`` as a local executable or a Docker image
            only.
        image_format
            Format of the image file(s) for the exported diagram(s).
            This can be set to any value out of ``"bmp"``, ``"gif"``,
            ``"jpg"``, ``"png"``, or ``"svg"``.
        create_index
            If ``True``, two index files ``index.json`` and
            ``index.html`` will be created. The JSON file consists of a
            list of dictionaries, each representing a diagram in the
            model. The dictionaries come with the keys

            - uuid: The unique ID of the diagram
            - name: Name of the diagram as it has been set in Capella
            - type: The diagram type as it was created in Capella
            - viewpoint: The source layer from where the representation
              is loaded from. It is ``Common`` for layerless diagrams.
            - success: A boolean stating if a diagram has been exported
              from Capella

            The HTML file shows a numbered list of diagram names which
            are hyperlinked to the diagram image file. Right beside a
            diagram's name one can also see the diagram's UUID in a
            discreet light gray and tiny font size. The HTML index also
            provides some meta data like a timestamp for the
            update of diagrams.
        force
            If the value of ``capella_cli`` is ambiguous and can match
            both a local executable and a Docker image, this parameter
            can be used to bypass the auto-detection and force the
            choice. A value of ``"exe"`` always interprets
            ``capella_cli`` as local executable, ``"docker"`` always
            interprets it as a docker image name. ``None`` (the default)
            enables automatic detection.
        background
            Add a white background to exported SVG images.

            Ignored if the ``image_format`` is not ``"svg"``.

        Examples
        --------
        Passing a bare filename looks up the executable in the PATH,
        after replacing a possible '{VERSION}' field:

        >>> model.update_diagram_cache("capella", "png")
        >>> model.update_diagram_cache("capella{VERSION}", "png")

        Passing an absolute path to a local installation of Capella that
        contains the Capella version will use that executable:

        >>> model.update_diagram_cache("/opt/capella{VERSION}/capella", "png")

        Passing a docker image name will launch a docker container, using the
        Capella binary at the image's ENTRYPOINT:

        >>> model.update_diagram_cache(
        ...     "ghcr.io/dbinfrago/capella-dockerimages/capella/base:{VERSION}-selected-dropins-main",
        ...     "png",
        ... )
        """
        if self.diagram_cache is None:
            raise TypeError(
                "Cannot update: No diagram_cache was specified for this model"
            )
        _diagram_cache.export(
            capella_cli,
            self,
            format=image_format,
            index=create_index,
            force=force,
            background=background,
        )

    @property
    def info(self) -> ModelInfo:
        upstream_info = self._loader.get_model_info()
        if self.diagram_cache is not None:
            dgcinfo = self.diagram_cache.get_model_info()
        else:
            dgcinfo = None
        return ModelInfo(
            url=upstream_info.url,
            title=upstream_info.title,
            entrypoint=upstream_info.entrypoint,
            resources=upstream_info.resources,
            capella_version=upstream_info.capella_version,
            viewpoints=upstream_info.viewpoints,
            diagram_cache=dgcinfo,
        )

    @property
    def description_badge(self) -> str:
        """Describe model contents distribution with an SVG badge."""
        from capellambse.extensions import metrics  # noqa: PLC0415

        return metrics.get_summary_badge(self)

    def referenced_viewpoints(self) -> dict[str, str]:
        return self.__viewpoints.copy()

    def activate_viewpoint(self, name: str, version: str) -> None:
        self.__viewpoints[name] = version
        self._loader.activate_viewpoint(name, version)

    def resolve_class(
        self,
        typehint: (
            str | _obj.UnresolvedClassName | etree.QName | etree._Element
        ),
    ) -> type[_obj.ModelObject]:
        """Resolve a class based on a type hint.

        The type hint can be any one of:

        - A string like ``'namespace.alias:ClassName'``, as used in the
          ``xsi:type`` XML attribute
        - A string like ``'{http://name/space/url}ClassName'``, as used
          in XML tags for fragment roots. For versioned namespaces,
          the version included in the URL is ignored. Future versions
          may raise an error if the version doesn't match the activated
          viewpoint.
        - A :class:`~lxml.etree.QName` object, as obtained by
          :func:`~capellambse.helpers.qtype_of`.
        - An :class:`~lxml.etree._Element` object.
        - A simple ClassName string, which will be searched across all
          namespaces. It is an error if multiple namespaces provide a
          class with that name; to avoid such errors, always use a form
          that explicitly provides the namespace.
        - A tuple of ``('name.space.alias', 'ClassName')``
        - A tuple of ``(NamespaceInstance, 'ClassName')``
        """
        if isinstance(typehint, etree._Element):
            qtype = capellambse.helpers.qtype_of(typehint)
            if qtype is None:
                raise ValueError(
                    f"Element is not a proper model element: {typehint!r}"
                )
            typehint = qtype

        if isinstance(typehint, etree.QName):
            ns, _ = _obj.find_namespace_by_uri(typehint.namespace)
            clsname = typehint.localname

        elif isinstance(typehint, str):
            if typehint.startswith("{"):
                qn = etree.QName(typehint)
                ns, _ = _obj.find_namespace_by_uri(qn.namespace)
                clsname = qn.localname
            elif ":" in typehint:
                nsalias, clsname = typehint.rsplit(":", 1)
                ns = _obj.find_namespace(nsalias)
            else:
                clsname = typehint
                providers = [
                    i for i in _obj.enumerate_namespaces() if clsname in i
                ]
                if len(providers) < 1:
                    raise _obj.MissingClassError(None, None, clsname)
                if len(providers) > 1:
                    raise ValueError(
                        f"Multiple namespaces providing class {clsname!r}:"
                        f" {', '.join(i.alias for i in providers)}"
                    )
                (ns,) = providers

        elif isinstance(typehint, tuple) and len(typehint) == 2:
            clsname = typehint[1]
            if isinstance(typehint[0], str):
                ns = _obj.find_namespace(typehint[0])
            else:
                ns = typehint[0]

        else:
            raise TypeError(
                f"Invalid typehint, expected a str or 2-tuple: {typehint!r}"
            )

        if ns.viewpoint is None or "{VERSION}" not in ns.uri:
            return ns.get_class(clsname)
        viewpoint = self.referenced_viewpoints().get(ns.viewpoint)
        if viewpoint is None:
            raise RuntimeError(
                f"Required viewpoint not activated: {viewpoint!r}"
            )
        return ns.get_class(clsname, viewpoint)

    def qualify_classname(self, cls: _obj.ClassName, /) -> etree.QName:
        """Qualify a ClassName based on the activated viewpoints."""
        ns, clsname = cls
        if "{VERSION}" not in ns.uri or not ns.viewpoint:
            return etree.QName(ns.uri, clsname)

        vp = self.referenced_viewpoints().get(ns.viewpoint)
        if vp is None:
            raise _descriptors.InvalidModificationError(
                f"Required viewpoint is not activated: {ns.viewpoint}"
            )
        vp = ns.trim_version(vp)
        return etree.QName(ns.uri.format(VERSION=vp), clsname)

    if t.TYPE_CHECKING:

        def __getattr__(self, attr: str) -> t.Any:
            """Account for extension attributes in static type checks."""

        def __setattr__(
            self,
            attr: str,
            val: property | cabc.Callable[..., t.Any],
        ) -> None:
            """Allow setting additional properties in static type checks."""


@dataclasses.dataclass
class ModelInfo(loader.ModelInfo):
    diagram_cache: filehandler.abc.HandlerInfo | None


def _reference_attributes(
    objtype: type[_obj.ModelObject], /
) -> tuple[str, ...]:
    ignored_accessors: tuple[type[_descriptors.Accessor], ...] = (
        _descriptors.AlternateAccessor,
        _descriptors.DeepProxyAccessor,
        _descriptors.DeprecatedAccessor,
        _descriptors.ParentAccessor,
        _descriptors.Backref,
    )

    attrs: list[str] = []
    for i in dir(objtype):
        if i.startswith("_") or i == "parent":
            continue
        acc = getattr(objtype, i, None)
        if (
            isinstance(acc, _descriptors.Relationship)
            or (
                isinstance(acc, _descriptors.Single)
                and isinstance(acc.wrapped, _descriptors.Relationship)
            )
            # TODO remove checks for deprecated Accessor classes
            or (
                isinstance(acc, _descriptors.Accessor)
                and not isinstance(acc, ignored_accessors)
            )
        ):
            attrs.append(i)
    return tuple(attrs)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Static duck-typing assertions checked by mypy."""

from __future__ import annotations

from capellambse import model


def protocol_ModelObject_compliance():
    mobj: model.ModelObject

    mobj = model.ModelElement()  # type: ignore[call-arg]
    mobj = model._descriptors._Specification()  # type: ignore[call-arg]
    mobj = model.diagram.Diagram()

    del mobj



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Global fixtures for pytest."""

import importlib.metadata as imm
import pathlib
import shutil

import pytest

import capellambse

INSTALLED_PACKAGE = pathlib.Path(capellambse.__file__).parent
TEST_DATA = pathlib.Path(__file__).parent / "data"

capellambse.load_model_extensions()

if not any(
    i.module.startswith("capellambse.extensions.")
    for i in imm.entry_points(group="capellambse.model_extensions")
):
    raise RuntimeError(
        "Built-in model extensions are not loaded,"
        " is capellambse installed properly?"
    )


class Models:
    aird_parser = TEST_DATA.joinpath("models", "aird_parser")
    empty = TEST_DATA.joinpath("models", "empty")
    filtering = TEST_DATA.joinpath("models", "filtering")
    lib_proj = TEST_DATA.joinpath("models", "library_project")
    lib_test = TEST_DATA.joinpath("models", "library_test")
    pvmt = TEST_DATA.joinpath("models", "pvmt")
    test7_0 = TEST_DATA.joinpath("models", "test7_0")
    writemodel = TEST_DATA.joinpath("models", "writemodel")


@pytest.fixture(scope="session")
def session_shared_model() -> capellambse.MelodyModel:
    """Load the standard test model.

    Unlike the ``model`` fixture, this fixture is shared across the
    entire test session. As such, the test functions using this fixture
    are expected to not modify it.

    This fixture exists as a speed optimization for tests that only read
    from the model.
    """
    return capellambse.MelodyModel(Models.test7_0)


@pytest.fixture
def model() -> capellambse.MelodyModel:
    """Load the Capella 7.0 test model."""
    return capellambse.MelodyModel(Models.test7_0)


@pytest.fixture
def empty_model() -> capellambse.MelodyModel:
    """Load the empty test model."""
    return capellambse.MelodyModel(Models.empty)


@pytest.fixture
def tmp_model(tmp_path: pathlib.Path) -> pathlib.Path:
    """Copy the Capella 7.0 test model to a temporary directory."""
    path = tmp_path / "model"
    shutil.copytree(Models.test7_0, path)
    return path



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Support for YAML-based declarative modelling.

A YAML-based approach to describing how to create and modify
``capellambse`` compatible models.

For an in-depth explanation, please refer to the :ref:`full
documentation about declarative modelling <declarative-modelling>`.
"""

from __future__ import annotations

__all__ = [
    "FindBy",
    "Metadata",
    "ModelMetadata",
    "Promise",
    "UUIDReference",
    "UnfulfilledPromisesError",
    "WriterMetadata",
    "YDMDumper",
    "YDMLoader",
    "apply",
    "dump",
    "load",
    "load_with_metadata",
]

import collections
import collections.abc as cabc
import contextlib
import dataclasses
import importlib.metadata as imm
import logging
import operator
import os
import pathlib
import re
import subprocess
import sys
import typing as t
import warnings

import awesomeversion as av
import markupsafe
import typing_extensions as te
import yaml

import capellambse
import capellambse.model as m
from capellambse import helpers
from capellambse.model import NewObject

FileOrPath = str | os.PathLike[t.Any] | t.IO[str]
_FutureAction = dict[str, t.Any]
_OperatorResult = tuple[
    "Promise",
    capellambse.ModelObject | _FutureAction,
]
logger = logging.getLogger(__name__)


class WriterMetadata(t.TypedDict):
    capellambse: str
    generator: te.NotRequired[str]


class ModelMetadata(t.TypedDict):
    url: str
    revision: te.NotRequired[str]
    entrypoint: str


class Metadata(t.TypedDict, total=False):
    written_by: WriterMetadata
    model: ModelMetadata


@t.overload
def dump(
    instructions: cabc.Sequence[cabc.Mapping[str, t.Any]],
    *,
    metadata: Metadata | None = None,
) -> str: ...
@t.overload
def dump(
    instructions: cabc.Sequence[cabc.Mapping[str, t.Any]],
    *,
    metadata: m.MelodyModel,
    generator: str | None = None,
) -> str: ...
def dump(
    instructions: cabc.Sequence[cabc.Mapping[str, t.Any]],
    *,
    metadata: m.MelodyModel | Metadata | None = None,
    generator: str | None = None,
) -> str:
    """Dump an instruction stream to YAML.

    Optionally dump metadata with the instruction stream to YAML.
    """
    if isinstance(metadata, m.MelodyModel):
        res_info = metadata.info.resources["\x00"]
        metadata = {
            "model": {
                "url": res_info.url,
                "revision": res_info.rev_hash,
                "entrypoint": str(metadata.info.entrypoint),
            },
            "written_by": {
                "capellambse": capellambse.__version__.split("+", 1)[0]
            },
        }
        if generator is not None:
            metadata["written_by"]["generator"] = generator

    if not metadata:
        return yaml.dump(instructions, Dumper=YDMDumper)
    return yaml.dump_all([metadata, instructions], Dumper=YDMDumper)


def load(file: FileOrPath) -> list[dict[str, t.Any]]:
    """Load an instruction stream from a YAML file.

    Parameters
    ----------
    file
        An open file-like object containing decl instructions, or a path
        or PathLike pointing to such a file. Files are expected to use
        UTF-8 encoding.
    """
    _, instructions = load_with_metadata(file)
    return instructions


def load_with_metadata(
    file: FileOrPath,
) -> tuple[Metadata, list[dict[str, t.Any]]]:
    """Load an instruction stream and its metadata from a YAML file.

    If the file does not have a metadata section, an empty dict will be
    returned.

    Parameters
    ----------
    file
        An open file-like object containing decl instructions, or a path
        or PathLike pointing to such a file. Files are expected to use
        UTF-8 encoding.

    Returns
    -------
    dict[str, Any]
        The metadata read from the file, or an empty dictionary if the
        file did not contain any metadata.
    list[dict[str, Any]]
        The instruction stream.
    """
    if hasattr(file, "read"):
        file = t.cast(t.IO[str], file)
        ctx: t.ContextManager[t.IO[str]] = contextlib.nullcontext(file)
    else:
        assert not isinstance(file, t.IO)
        ctx = open(file, encoding="utf-8")  # noqa: SIM115

    with ctx as opened_file:
        contents = list(yaml.load_all(opened_file, Loader=YDMLoader))

    if len(contents) == 2:
        return (t.cast(Metadata, contents[0]) or {}, contents[1] or [])
    if len(contents) == 1:
        return ({}, contents[0] or [])
    if len(contents) == 0:
        return ({}, [])
    raise ValueError(
        f"Expected a YAML file with 1 or 2 documents, found {len(contents)}"
    )


def apply(
    model: capellambse.MelodyModel,
    file: FileOrPath,
    *,
    strict: bool = False,
) -> dict[Promise, capellambse.ModelObject]:
    """Apply a declarative modelling file to the given model.

    Parameters
    ----------
    model
        The model to apply the instructions to.
    file
        An open file-like object to read YAML instructions from, or a
        path to such a file. Files will be read with UTF-8 encoding.

        The full format of these files is documented in the
        :ref:`section about declarative modelling
        <declarative-modelling>`.
    strict
        Verify metadata contained in the file against the used model,
        and raise an error if they don't match.

    Notes
    -----
    This function is not transactional: If an exception occurs during
    this function, the model will be left partially modified, with no
    reliable way to know how much of the YAML input has been consumed.
    It is therefore advised to reload or discard the model immediately
    in these cases, to avoid working with an inconsistent state.

    Even though the YAML layout suggests linear execution from top to
    bottom, the actual order in which modifications are executed is
    implementation defined. This is necessary to support promises with
    ``!promise``, but reorderings are still possible even if no promises
    are used in an input document.
    """
    metadata, raw_instructions = load_with_metadata(file)
    instructions = collections.deque(raw_instructions)
    promises = dict[Promise, capellambse.ModelObject]()
    deferred = collections.defaultdict[Promise, list[_FutureAction]](list)

    if strict:
        if not metadata:
            raise ValueError("No metadata found to verify in strict mode")
        try:
            _verify_metadata(model, metadata)
        except TypeError as err:
            raise TypeError(f"Cannot apply decl: {err}") from None
    elif metadata:
        try:
            _verify_metadata(model, metadata)
        except ValueError as err:
            logger.warning("Metadata does not match provided model: %s", err)

    while instructions:
        instruction = instructions.popleft()

        parent = instruction.pop("parent")
        if isinstance(parent, Promise | _ObjectFinder):
            try:
                parent = _resolve(promises, model.project, parent)
            except _UnresolvablePromise as p:
                d = {"parent": parent, **instruction}
                deferred[p.args[0]].append(d)
                continue

        if not isinstance(parent, capellambse.model.ModelElement):
            raise TypeError(
                "Expected a model object as parent, found "
                f"{type(parent).__name__}"
            )

        for op_type, apply_op in _OPERATIONS.items():
            try:
                op = instruction.pop(op_type)
            except KeyError:
                continue
            for promise, outcome in apply_op(promises, parent, op):
                if isinstance(outcome, dict):
                    deferred[promise].append(outcome)
                else:
                    if promise in promises:
                        raise ValueError(
                            f"promise_id defined twice: {promise.identifier}"
                        )
                    promises[promise] = outcome
                    instructions.extend(deferred.pop(promise, ()))
        if instruction:
            keys = ", ".join(instruction)
            raise ValueError(f"Unrecognized keys in instruction: {keys}")
    if deferred:
        raise UnfulfilledPromisesError(frozenset(deferred))
    return promises


def _verify_metadata(
    model: capellambse.MelodyModel, metadata: Metadata
) -> None:
    assert metadata

    written_by = metadata.get("written_by", {}).get("capellambse", "")
    if not written_by:
        raise ValueError("Can't find 'written_by:capellambse' in metadata")
    if not _is_pep440(written_by):
        raise ValueError(f"Malformed version number in metadata: {written_by}")

    current = av.AwesomeVersion(
        imm.version("capellambse").partition("+")[0],
        ensure_strategy=av.AwesomeVersionStrategy.PEP440,
    )
    try:
        written_version = av.AwesomeVersion(
            written_by,
            ensure_strategy=av.AwesomeVersionStrategy.PEP440,
        )
        version_matches = current >= written_version
    except Exception as err:
        raise ValueError(
            "Cannot verify required capellambse version:"
            f" {type(err).__name__}: {err}"
        ) from None

    if not version_matches:
        raise ValueError(
            "This capellambse is too old for this YAML:"
            f" Need at least v{written_by}, but have only v{current})"
        )

    model_metadata = metadata.get("model", {})
    res_info = model.info.resources["\x00"]
    url = model_metadata.get("url")
    if url != res_info.url:
        raise ValueError(
            "Model URL mismatch:"
            f" YAML expects {url}, current is {res_info.url}"
        )

    hash = model_metadata.get("revision")
    if hash != res_info.rev_hash:
        raise ValueError(
            "Model version mismatch:"
            f" YAML expects {hash}, current is {res_info.rev_hash}"
        )

    entrypoint = pathlib.PurePosixPath(model_metadata.get("entrypoint", ""))
    if entrypoint != model.info.entrypoint:
        raise ValueError(
            "Model entrypoint mismatch:"
            f" YAML expects {entrypoint}, current is {model.info.entrypoint}"
        )


def _is_pep440(version: str) -> bool:
    """Check if given version aligns with PEP440.

    See Also
    --------
    https://peps.python.org/pep-0440/#appendix-b-parsing-version-strings-with-regular-expressions
    """
    pep440_ptrn = re.compile(
        r"([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)"
        r"(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?"
    )
    return pep440_ptrn.fullmatch(version) is not None


def _operate_create(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    creations: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    yield from _operate_extend(promises, parent, creations)


def _operate_extend(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    extensions: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    for attr, value in extensions.items():
        if not isinstance(value, cabc.Iterable):
            raise TypeError("values below `extend:*:` must be lists")

        yield from _create_complex_objects(promises, parent, attr, value)


def _operate_delete(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    deletions: dict[str, t.Any],
) -> cabc.Iterable[_OperatorResult]:
    del promises

    for attr, objs in deletions.items():
        try:
            target = getattr(parent, attr)
        except AttributeError:
            raise TypeError(
                "Cannot delete object:"
                f" {type(parent).__name__} has no attribute {attr!r}"
            ) from None
        if not isinstance(target, m.ElementList) or not isinstance(objs, list):
            delattr(parent, attr)
            continue
        if not isinstance(target, m.ElementListCouplingMixin):
            raise TypeError(
                "Cannot delete object:"
                f" {type(parent).__name__}.{attr} is not model-coupled"
            )
        for obj in objs:
            if isinstance(obj, Promise):
                raise ValueError("Cannot use !promise in `delete:*:`")
            if isinstance(obj, str):
                obj = UUIDReference(helpers.UUIDString(obj))
            obj = _resolve({}, parent, obj)
            try:
                idx = target.index(obj)
            except ValueError:
                if hasattr(parent, "_short_repr_"):
                    p_repr = parent._short_repr_()
                else:
                    p_repr = repr(getattr(parent, "uuid", "<unknown>"))
                raise ValueError(
                    f"No object {obj._short_repr_()} in {attr!r} of {p_repr}"
                ) from None
            del target[idx]

    return ()


def _operate_set(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    modifications: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    for attr, value in modifications.items():
        if isinstance(value, list | Promise | _ObjectFinder):
            try:
                value = _resolve(promises, parent, value)
            except _UnresolvablePromise as p:
                yield p.args[0], {"parent": parent, "set": {attr: value}}
                continue

        if isinstance(value, list):
            getattr(parent, attr).clear()
            yield from _create_complex_objects(promises, parent, attr, value)
        elif isinstance(value, dict):
            obj = getattr(parent, attr)
            for k, v in value.items():
                setattr(obj, k, v)
        else:
            setattr(parent, attr, value)


def _operate_sync(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    modifications: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    for attr, value in modifications.items():
        if not isinstance(value, cabc.Iterable):
            raise TypeError("values below `extend:*:` must be lists")

        for obj in value:
            try:
                find_args = obj["find"]
            except KeyError:
                raise ValueError(
                    "Expected `find` key in sync object"
                ) from None
            if isinstance(find_args, dict):
                find_args = FindBy(find_args)
            assert isinstance(find_args, FindBy)

            try:
                candidate = _resolve_findby(promises, parent, attr, find_args)
            except _NoObjectFoundError:
                candidate = None
            except _UnresolvablePromise as p:
                yield p.args[0], {"parent": parent, "sync": {attr: [obj]}}
                continue

            if candidate is not None:
                if sync := obj.pop("sync", None):
                    yield from _operate_sync(promises, candidate, sync)
                if mods := obj.pop("set", None):
                    yield from _operate_set(promises, candidate, mods)
                if ext := obj.pop("extend", None):
                    yield from _operate_extend(promises, candidate, ext)
                promise: str | Promise | None = obj.get("promise_id")
                if promise is not None:
                    if isinstance(promise, str):
                        promise = Promise(promise)
                    yield (promise, candidate)
            else:
                mods = obj.setdefault("set", {})
                ext = obj.setdefault("extend", {})
                try:
                    for k, v in mods.items():
                        mods[k] = _resolve(promises, parent, v)
                    for k, v in ext.items():
                        ext[k] = _resolve(promises, parent, v)
                except _UnresolvablePromise as p:
                    deferred = {
                        "parent": parent,
                        "sync": {attr: [obj]},
                    }
                    yield p.args[0], deferred
                else:
                    newobj_props = find_args.attributes | mods | ext
                    if "promise_id" in obj:
                        newobj_props["promise_id"] = obj.pop("promise_id")
                    yield from _create_complex_objects(
                        promises, parent, attr, [newobj_props]
                    )
                    if "sync" in obj:
                        yield from _operate_sync(
                            promises, parent, {attr: [obj]}
                        )


def _resolve(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    value: t.Any,
) -> t.Any:
    if isinstance(value, Promise):
        try:
            return promises[value]
        except KeyError:
            raise _UnresolvablePromise(value) from None
    elif isinstance(value, UUIDReference):
        return parent._model.by_uuid(value.uuid)
    elif isinstance(value, FindBy):
        return _resolve_findby(promises, parent, None, value)
    elif isinstance(value, list):
        for i, v in enumerate(value):
            newv = _resolve(promises, parent, v)
            if newv is not v:
                value[i] = newv
    return value


def _resolve_findby(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    attr: str | None,
    value: FindBy,
) -> capellambse.ModelObject:
    attrs = dict(value.attributes)
    typehint = attrs.pop("_type", None)
    if not isinstance(typehint, str | type(None)):
        raise TypeError(
            f"Expected a string for !find {{_type: ...}},"
            f" got {type(typehint)}: {typehint!r}"
        )
    if typehint is None:
        wanted_types: tuple[type[t.Any], ...] = ()
    else:
        wanted_types = (parent._model.resolve_class(typehint),)

    if isinstance(parent, capellambse.MelodyModel):
        candidates = parent.search(*wanted_types)
    elif attr is not None:
        candidates = getattr(parent, attr)
        if wanted_types:
            candidates = candidates.filter(
                lambda i: isinstance(i, wanted_types)
            )
    else:
        candidates = parent._model.search(*wanted_types)

    if attrs:
        for k, v in attrs.items():
            if isinstance(v, list | Promise | _ObjectFinder):
                attrs[k] = _resolve(promises, parent, v)

        if len(attrs) > 1:
            expected_values = tuple(attrs.values())
        else:
            (expected_values,) = attrs.values()
        getter = operator.attrgetter(*attrs)

        def do_filter(obj):
            try:
                real_values = getter(obj)
            except AttributeError:
                return False
            return real_values == expected_values

        candidates = candidates.filter(do_filter)

    if len(candidates) > 1:
        hint = "(Hint: did you mean '_type' instead of 'type'?)\n" * (
            "type" in value.attributes and "_type" not in value.attributes
        )
        raise ValueError(
            f"Ambiguous match directive: !find {value.attributes!r}\n"
            + hint
            + f"Found {len(candidates)} matches:\n"
            + candidates._short_repr_()
        )
    if not candidates:
        raise _NoObjectFoundError(value.attributes)
    return candidates[0]


class _UnresolvablePromise(BaseException):
    pass


class _NoObjectFoundError(ValueError):
    def __str__(self) -> str:
        if len(self.args) == 1:
            return f"No object found for !find {self.args[0]!r}"
        return super().__str__()


_OPERATIONS = collections.OrderedDict(
    (
        ("create", _operate_create),
        ("extend", _operate_extend),
        ("set", _operate_set),
        ("sync", _operate_sync),
        ("delete", _operate_delete),
    )
)


def _create_complex_objects(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    attr: str,
    objs: cabc.Iterable[dict[str, t.Any] | Promise | str],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    try:
        target = getattr(parent, attr)
    except AttributeError:
        raise TypeError(
            "Cannot create object:"
            f" {type(parent).__name__} has no attribute {attr!r}"
        ) from None
    if not isinstance(target, m.ElementList):
        raise TypeError(
            "Cannot create object:"
            f" {type(parent).__name__}.{attr} is not a list"
        )
    if not isinstance(target, m.ElementListCouplingMixin):
        raise TypeError(
            "Cannot create object:"
            f" {type(parent).__name__}.{attr} is not model-coupled"
        )
    for child in objs:
        if isinstance(
            child,
            m.ModelElement | list | Promise | _ObjectFinder,
        ):
            try:
                obj = _resolve(promises, parent, child)
            except _UnresolvablePromise as p:
                yield p.args[0], {"parent": parent, "extend": {attr: [child]}}
            else:
                target.append(obj)
        elif isinstance(child, str):
            target.create_singleattr(child)
        else:
            assert not isinstance(child, Promise)  # mypy false-positive
            yield from _create_complex_object(
                promises, parent, attr, target, child
            )


def _create_complex_object(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    attr: str,
    target: m.ElementListCouplingMixin,
    obj_desc: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    try:
        promise: str | Promise | None = obj_desc.pop("promise_id")
    except KeyError:
        promise = None

    complex_attrs = dict[str, t.Any]()
    simple_attrs = dict[str, t.Any]()
    type_hint = tuple[str, ...]()
    try:
        for k, v in obj_desc.items():
            if k == "_type":
                type_hint = (v,)
            elif isinstance(v, cabc.Iterable) and not isinstance(v, str):
                complex_attrs[k] = v
            else:
                simple_attrs[k] = _resolve(promises, parent, v)
    except _UnresolvablePromise as p:
        obj_desc["promise_id"] = promise
        yield (p.args[0], {"parent": parent, "extend": {attr: [obj_desc]}})
        return
    assert isinstance(target, m.ElementListCouplingMixin)
    obj = target.create(*type_hint, **simple_attrs)
    if promise is not None:
        if isinstance(promise, str):
            promise = Promise(promise)
        yield (promise, obj)
    for subkey, subval in complex_attrs.items():
        if isinstance(subval, cabc.Iterable):
            yield from _create_complex_objects(promises, obj, subkey, subval)
        else:
            raise TypeError(f"Expected list, got {type(subval).__name__}")


@dataclasses.dataclass(frozen=True)
class Promise:
    """References a model object that will be created later."""

    identifier: str


class UnfulfilledPromisesError(RuntimeError):
    """A promise could not be fulfilled.

    This exception is raised when a promise is referenced via
    ``!promise``, but it is never fulfilled by declaring an object with
    the same ``promise_id``.
    """

    def __str__(self) -> str:
        if (
            len(self.args) == 1
            and isinstance(self.args[0], cabc.Iterable)
            and not isinstance(self.args[0], str)
        ):
            return ", ".join(i.identifier for i in self.args[0])
        return super().__str__()


@dataclasses.dataclass(frozen=True)
class UUIDReference:
    """References a model object by its UUID."""

    uuid: helpers.UUIDString

    def __post_init__(self) -> None:
        if not helpers.is_uuid_string(self.uuid):
            raise ValueError(f"Malformed `!uuid`: {self.uuid!r}")


@dataclasses.dataclass(frozen=True)
class FindBy:
    """Find an object by specific attributes."""

    attributes: cabc.Mapping[str, t.Any]


_ObjectFinder: t.TypeAlias = UUIDReference | FindBy


class YDMDumper(yaml.SafeDumper):
    """A YAML dumper with extensions for declarative modelling."""

    def represent_promise(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, Promise)
        return self.represent_scalar("!promise", data.identifier)

    def represent_uuidref(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, UUIDReference)
        return self.represent_scalar("!uuid", data.uuid)

    def represent_newobj(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, NewObject)
        attrs = dict(data._kw)
        if data._type_hint:
            attrs["_type"] = data._type_hint
        return self.represent_mapping("!new_object", attrs)

    def represent_findby(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, FindBy)
        attrs = dict(data.attributes)
        return self.represent_mapping("!find", attrs)

    def represent_markup(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, markupsafe.Markup)
        return self.represent_scalar("!html", str(data))


YDMDumper.add_representer(Promise, YDMDumper.represent_promise)
YDMDumper.add_representer(UUIDReference, YDMDumper.represent_uuidref)
YDMDumper.add_representer(NewObject, YDMDumper.represent_newobj)
YDMDumper.add_representer(FindBy, YDMDumper.represent_findby)
YDMDumper.add_representer(markupsafe.Markup, YDMDumper.represent_markup)


class YDMLoader(yaml.SafeLoader):
    """A YAML loader with extensions for declarative modelling."""

    def construct_promise(self, node: yaml.Node) -> Promise:
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError("!promise only accepts scalar nodes")
        data = self.construct_scalar(node)
        if not isinstance(data, str):
            raise TypeError("!promise only accepts string scalars")
        return Promise(data)

    def construct_uuidref(self, node: yaml.Node) -> UUIDReference:
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError("!uuid only accepts scalar nodes")
        data = self.construct_scalar(node)
        if not helpers.is_uuid_string(data):
            raise ValueError(f"Malformed UUID: {data}")
        return UUIDReference(data)

    def construct_newobj(self, node: yaml.Node) -> NewObject:
        if not isinstance(node, yaml.MappingNode):
            raise TypeError("!new_object only accepts mapping nodes")
        data = self.construct_mapping(node)
        try:
            _type = data.pop("_type")
        except KeyError:
            raise ValueError("!new_object requires a _type key") from None
        return NewObject(_type, **t.cast(t.Any, data))

    def construct_findby(self, node: yaml.Node) -> FindBy:
        if not isinstance(node, yaml.MappingNode):
            raise TypeError("!find only accepts mapping nodes")
        data = self.construct_mapping(node)
        return FindBy(t.cast(t.Any, data))

    def construct_markup(self, node: yaml.Node) -> markupsafe.Markup:
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError("!html only accepts scalar nodes")
        return markupsafe.Markup(self.construct_scalar(node))


YDMLoader.add_constructor("!promise", YDMLoader.construct_promise)
YDMLoader.add_constructor("!uuid", YDMLoader.construct_uuidref)
YDMLoader.add_constructor("!new_object", YDMLoader.construct_newobj)
YDMLoader.add_constructor("!find", YDMLoader.construct_findby)
YDMLoader.add_constructor("!html", YDMLoader.construct_markup)


def _main() -> None:
    name = "decl-apply"
    warnings.warn(
        f"This CLI entry point is deprecated, use 'capellambse {name}' instead",
        FutureWarning,
        stacklevel=2,
    )

    cmd = [sys.executable, "-mcapellambse", name, *sys.argv[1:]]
    raise SystemExit(subprocess.run(cmd, check=False).returncode)


if __name__ == "__main__":
    _main()



#!/usr/bin/env -S python -Xdev -Xfrozen_modules=off
# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import subprocess
import sys
import warnings


def main() -> None:
    name = "repl"
    warnings.warn(
        f"This CLI entry point is deprecated, use 'capellambse {name}' instead",
        FutureWarning,
        stacklevel=2,
    )

    cmd = [
        sys.executable,
        "-Xdev",
        "-Xfrozen_modules=off",
        "-mcapellambse",
        name,
        *sys.argv[1:],
    ]
    raise SystemExit(subprocess.run(cmd, check=False).returncode)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import logging
import typing as t

import click
import jinja2

import capellambse


@click.command()
@click.option("-m", "--model", required=True, type=capellambse.ModelCLI())
@click.option(
    "-o",
    "--output",
    type=click.File("wb", atomic=True),
    required=True,
    help="Output file to render the template into",
)
@click.option("-t", "--template", help="An optional custom template to render")
def main(
    model: capellambse.MelodyModel,
    template: str | None,
    output: t.IO[bytes],
) -> None:
    """Validate a model against predefined rules."""
    logging.basicConfig()

    loader: jinja2.BaseLoader
    if template is None:
        loader = jinja2.PackageLoader("capellambse", "extensions/validation")
        template = "report-template.html.jinja"
    else:
        loader = jinja2.FileSystemLoader(".")
    env = jinja2.Environment(loader=loader)

    with output:
        env.get_template(template).stream(
            model=model,
            results=model.validation.validate(),
        ).dump(output, encoding="utf-8")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Logical Architecture layer."""

from __future__ import annotations

import sys
import typing as t

import capellambse.model as m

from . import capellacommon, cs, fa, interaction
from . import namespaces as ns

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

NS = ns.LA


class LogicalArchitecturePkg(cs.BlockArchitecturePkg):
    architectures = m.Containment["LogicalArchitecture"](
        "ownedLogicalArchitectures", (NS, "LogicalArchitecture")
    )


class LogicalArchitecture(cs.ComponentArchitecture):
    component_pkg = m.Single["LogicalComponentPkg"](
        m.Containment("ownedLogicalComponentPkg", (NS, "LogicalComponentPkg"))
    )
    system_analysis_realizations = m.Containment["SystemAnalysisRealization"](
        "ownedSystemAnalysisRealizations", (NS, "SystemAnalysisRealization")
    )
    realized_system_analysis = m.Allocation["sa.SystemAnalysis"](
        "ownedSystemAnalysisRealizations",
        (NS, "SystemAnalysisRealization"),
        (ns.SA, "SystemAnalysis"),
        attr="targetElement",
        backattr="sourceElement",
    )

    @property
    def root_function(self) -> LogicalFunction:
        """Returns the first function in the function_pkg."""
        pkg = self.function_pkg
        assert pkg is not None
        if not pkg.functions:
            raise RuntimeError(f"Package {pkg._short_repr_()} is empty")
        return pkg.functions[0]

    @property
    def root_component(self) -> LogicalComponent:
        assert self.component_pkg is not None
        return self.component_pkg.components.by_is_actor(False, single=True)

    @property
    def all_components(self) -> m.ElementList[LogicalComponent]:
        return self._model.search((NS, "LogicalComponent"), below=self)

    @property
    def all_actors(self) -> m.ElementList[LogicalComponent]:
        return self._model.search(LogicalComponent).by_is_actor(True)

    @property
    def all_actor_exchanges(self) -> m.ElementList[fa.ComponentExchange]:
        return self._model.search(
            (ns.FA, "ComponentExchange"), below=self
        ).filter(
            lambda e: (
                (
                    e.source is not None
                    and e.source.parent is not None
                    and e.source.parent.is_actor
                )
                or (
                    e.target is not None
                    and e.target.parent is not None
                    and e.target.parent.is_actor
                )
            )
        )

    @property
    def all_component_exchanges(self) -> m.ElementList[fa.ComponentExchange]:
        return self._model.search((ns.FA, "ComponentExchange"), below=self)

    @property
    @deprecated(
        (
            "LogicalArchitecture.component_exchanges will soon change"
            " to map the directly contained ComponentExchanges;"
            " use 'all_component_exchanges' to refer to"
            " all recursively contained exchanges instead"
        ),
        category=FutureWarning,
    )
    def component_exchanges(self) -> m.ElementList[fa.ComponentExchange]:  # type: ignore[override]
        return self.all_component_exchanges

    diagrams = m.DiagramAccessor(
        "Logical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )

    if not t.TYPE_CHECKING:
        component_package = m.DeprecatedAccessor("component_pkg")
        actor_exchanges = m.DeprecatedAccessor("all_actor_exchanges")


class LogicalFunction(fa.AbstractFunction):
    functions = m.Containment["LogicalFunction"](
        "ownedFunctions", (NS, "LogicalFunction")
    )
    packages = m.Containment["LogicalFunctionPkg"](
        "ownedLogicalFunctionPkgs", (NS, "LogicalFunctionPkg")
    )
    realized_system_functions = m.Alias["m.ElementList[sa.SystemFunction]"](
        "realized_functions"
    )
    owner = m.Single["LogicalComponent"](
        m.Backref((NS, "LogicalComponent"), "allocated_functions")
    )
    involved_in = m.Backref["CapabilityRealization"](
        (NS, "CapabilityRealization"), "involved_functions"
    )
    realizing_physical_functions = m.Backref["pa.PhysicalFunction"](
        (ns.PA, "PhysicalFunction"), "realized_logical_functions"
    )


class LogicalFunctionPkg(fa.FunctionPkg):
    _xmltag = "ownedFunctionPkg"

    functions = m.Containment["LogicalFunction"](
        "ownedLogicalFunctions", (NS, "LogicalFunction")
    )
    packages = m.Containment["LogicalFunctionPkg"](
        "ownedLogicalFunctionPkgs", (NS, "LogicalFunctionPkg")
    )


class LogicalComponent(
    cs.Component, capellacommon.CapabilityRealizationInvolvedElement
):
    _xmltag = "ownedLogicalComponents"

    components = m.Containment["LogicalComponent"](
        "ownedLogicalComponents", (NS, "LogicalComponent")
    )
    architectures = m.Containment["LogicalArchitecture"](
        "ownedLogicalArchitectures", (NS, "LogicalArchitecture")
    )
    packages = m.Containment["LogicalComponentPkg"](
        "ownedLogicalComponentPkgs", (NS, "LogicalComponentPkg")
    )
    realized_system_components = m.Alias["m.ElementList[sa.SystemComponent]"](
        "realized_components"
    )
    realizing_physical_components = m.Backref["pa.PhysicalComponent"](
        (ns.PA, "PhysicalComponent"), "realized_logical_components"
    )


class LogicalComponentPkg(cs.ComponentPkg):
    _xmltag = "ownedLogicalComponentPkg"

    components = m.Containment["LogicalComponent"](
        "ownedLogicalComponents", (NS, "LogicalComponent")
    )
    packages = m.Containment["LogicalComponentPkg"](
        "ownedLogicalComponentPkgs", (NS, "LogicalComponentPkg")
    )


class CapabilityRealization(interaction.AbstractCapability):
    _xmltag = "ownedCapabilityRealizations"

    involved_functions = m.Allocation[LogicalFunction](
        None, None, (NS, "LogicalFunction")
    )
    capability_realization_involvements = m.Containment[
        "capellacommon.CapabilityRealizationInvolvement"
    ](
        "ownedCapabilityRealizationInvolvements",
        (ns.CAPELLACOMMON, "CapabilityRealizationInvolvement"),
    )
    involved_elements = m.Allocation[
        "capellacommon.CapabilityRealizationInvolvedElement"
    ](
        "ownedCapabilityRealizationInvolvements",
        (ns.CAPELLACOMMON, "CapabilityRealizationInvolvement"),
        (ns.CAPELLACOMMON, "CapabilityRealizationInvolvedElement"),
        attr="involved",
    )
    involved_components = m.Filter["LogicalComponent"](
        "involved_elements", (NS, "LogicalComponent"), legacy_by_type=True
    )

    if not t.TYPE_CHECKING:
        owned_chains = m.DeprecatedAccessor("functional_chains")


class CapabilityRealizationPkg(capellacommon.AbstractCapabilityPkg):
    """A capability package that can hold capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = m.Containment["CapabilityRealization"](
        "ownedCapabilityRealizations", (NS, "CapabilityRealization")
    )
    packages = m.Containment["CapabilityRealizationPkg"](
        "ownedCapabilityRealizationPkgs", (NS, "CapabilityRealizationPkg")
    )


class SystemAnalysisRealization(cs.ArchitectureAllocation):
    pass


class ContextInterfaceRealization(cs.InterfaceAllocation):
    pass


from . import pa, sa  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import click

import capellambse
from capellambse import cli_helpers


@click.command()
def main() -> None:
    """Show information about known models."""
    print("User-defined 'known models' are searched in:")
    print()
    print(f"   {capellambse.dirs.user_config_path / 'known_models'}")
    print()
    print("The following models are currently known:")
    for file in cli_helpers.enumerate_known_models():
        name, _, _ = file.name.rpartition(".")
        print(f"  - {name}")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import hashlib
import io
import pathlib
import shutil
import subprocess
import sys
import textwrap

import pytest

import capellambse.metamodel as mm
import capellambse.model as m
from capellambse import decl, helpers
from capellambse.extensions import reqif
from capellambse.filehandler import git

from .conftest import INSTALLED_PACKAGE, TEST_DATA, Models  # type: ignore

DATAPATH = TEST_DATA / "decl"

ROOT_COMPONENT = helpers.UUIDString("0d2edb8f-fa34-4e73-89ec-fb9a63001440")
ROOT_FUNCTION = helpers.UUIDString("f28ec0f8-f3b3-43a0-8af7-79f194b29a2d")
TEACH_POTIONS_FUNC = helpers.UUIDString("83ba0220-54f2-48f7-bca1-cd87e39639f2")


class TestDumpLoad:
    @staticmethod
    def test_promises_are_serialized_with_promise_tag() -> None:
        id = "some-future-object"
        data = [{"parent": decl.Promise(id)}]
        expected = f"- parent: !promise {id!r}\n"

        actual = decl.dump(data)

        assert actual == expected

    @staticmethod
    def test_uuid_references_are_serialized_with_uuid_tag() -> None:
        uuid = helpers.UUIDString("00000000-0000-0000-0000-000000000000")
        data = [{"parent": decl.UUIDReference(uuid)}]
        expected = f"- parent: !uuid {uuid!r}\n"

        actual = decl.dump(data)

        assert actual == expected

    @staticmethod
    def test_promise_tags_are_deserialized_as_promise() -> None:
        id = "some-future-object"
        yaml = f"- parent: !promise {id!r}\n"
        expected = [{"parent": decl.Promise(id)}]

        actual = decl.load(io.StringIO(yaml))

        assert actual == expected

    @staticmethod
    def test_uuid_tags_are_deserialized_as_uuidreference() -> None:
        uuid = helpers.UUIDString("00000000-0000-0000-0000-000000000000")
        yaml = f"- parent: !uuid {uuid!r}\n"
        expected = [{"parent": decl.UUIDReference(uuid)}]

        actual = decl.load(io.StringIO(yaml))

        assert actual == expected

    @staticmethod
    def test_loading_with_metadata():
        yml = textwrap.dedent(
            """\
            written_by:
              capellambse_version: 1.0.0
              generator: Example 1.0.0
            model:
              version: "12345678"
              url: https://example.invalid
              entrypoint: path/to/model.aird
            ---
            []
            """
        )
        expected = {
            "written_by": {
                "capellambse_version": "1.0.0",
                "generator": "Example 1.0.0",
            },
            "model": {
                "url": "https://example.invalid",
                "version": "12345678",
                "entrypoint": "path/to/model.aird",
            },
        }

        actual, _ = decl.load_with_metadata(io.StringIO(yml))

        assert actual == expected


class TestApplyExtend:
    @staticmethod
    def test_decl_errors_on_unknown_operations(
        model: m.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              invalid_operation: {{}}
            """

        with pytest.raises(ValueError, match="invalid_operation"):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    @pytest.mark.parametrize(
        "parent_str",
        [
            pytest.param(f"!uuid {TEACH_POTIONS_FUNC}", id="!uuid"),
            pytest.param(
                "!find {_type: LogicalFunction, name: teach Potions}",
                id="!find",
            ),
        ],
    )
    def test_decl_finds_parent_to_act_on(
        model: m.MelodyModel, parent_str
    ) -> None:
        parent = model.by_uuid(TEACH_POTIONS_FUNC)
        funcname = "pass the unit test"
        yml = f"""\
            - parent: {parent_str}
              extend:
                functions:
                  - name: {funcname!r}
            """
        expected_len = len(model.search()) + 1
        assert funcname not in parent.functions.by_name

        decl.apply(model, io.StringIO(yml))

        actual_len = len(model.search())
        assert actual_len == expected_len
        assert funcname in parent.functions.by_name

    @staticmethod
    def test_decl_creates_each_object_in_a_list(
        model: m.MelodyModel,
    ) -> None:
        parent_obj = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the first test
                  - name: pass the second test
                  - name: pass the third test
            """
        expected_len = len(model.search()) + 3
        for i in ("first", "second", "third"):
            assert f"pass the {i} test" not in parent_obj.functions.by_name

        decl.apply(model, io.StringIO(yml))

        actual_len = len(model.search())
        assert actual_len == expected_len
        for i in ("first", "second", "third"):
            assert f"pass the {i} test" in parent_obj.functions.by_name

    @staticmethod
    def test_decl_creates_nested_complex_objects_where_they_belong(
        model: m.MelodyModel,
    ) -> None:
        root = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the unit test
                    functions:
                      - name: run the test function
                      - name: make assertions
            """
        expected_len = len(model.search()) + 3

        decl.apply(model, io.StringIO(yml))

        actual_len = len(model.search())
        assert actual_len == expected_len
        assert "pass the unit test" in root.functions.by_name
        parent = root.functions.by_name("pass the unit test", single=True)
        assert "run the test function" in parent.functions.by_name
        assert "make assertions" in parent.functions.by_name

    @staticmethod
    @pytest.mark.parametrize(
        "cls",
        [reqif.AttributeDefinition, reqif.AttributeDefinitionEnumeration],
    )
    def test_decl_can_disambiguate_creations_with_type_hints(
        model: m.MelodyModel, cls: type[m.ModelElement]
    ) -> None:
        module_id = "db47fca9-ddb6-4397-8d4b-e397e53d277e"
        module = model.by_uuid(module_id)
        yml = f"""\
            - parent: !uuid {module_id}
              extend:
                attributes:
                  - long_name: New attribute
                    _type: {cls.__name__}
            """
        assert "New attribute" not in module.attributes.by_long_name

        decl.apply(model, io.StringIO(yml))

        attrdefs = module.attributes
        assert "New attribute" in attrdefs.by_long_name
        newdef = attrdefs.by_long_name("New attribute", single=True)
        assert type(newdef) is cls

    @staticmethod
    def test_decl_can_create_simple_objects_by_passing_plain_strings(
        model: m.MelodyModel,
    ) -> None:
        attrdef_id = "637caf95-3229-4607-99a0-7d7b990bc97f"
        attrdef = model.by_uuid(attrdef_id)
        yml = f"""\
            - parent: !uuid {attrdef_id}
              extend:
                values:
                  - NewGame++
            """
        assert "NewGame++" not in attrdef.values

        decl.apply(model, io.StringIO(yml))

        assert "NewGame++" in attrdef.values

    @staticmethod
    def test_extend_operations_change_model_objects_in_place(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        fnc = model.by_uuid("8833d2dc-b862-4a50-b26c-6f7e0f17faef")
        old_parent = fnc.parent
        assert old_parent != root_function
        assert fnc not in root_function.functions
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - !uuid {fnc.uuid}
            """

        decl.apply(model, io.StringIO(yml))

        assert fnc in root_function.functions
        assert root_function == fnc.parent

    @staticmethod
    def test_extend_operations_with_promises_change_model_objects_in_place(
        model: m.MelodyModel,
    ) -> None:
        root_component = model.by_uuid(ROOT_COMPONENT)
        fnc_name = "The promised one"
        assert fnc_name not in root_component.allocated_functions.by_name
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: {fnc_name}
                    promise_id: promised-fnc
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise promised-fnc
            """

        decl.apply(model, io.StringIO(yml))

        assert fnc_name in root_component.allocated_functions.by_name

    @staticmethod
    def test_extend_operations_on_faulty_attribute_cause_an_exception(
        model: m.MelodyModel,
    ) -> None:
        non_existing_attr = "Thought up"
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                {non_existing_attr}:
                  - !uuid {ROOT_FUNCTION}
            """

        with pytest.raises(TypeError, match=non_existing_attr):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    def test_promises_are_resolved_during_the_second_attempt(
        model: m.MelodyModel,
    ) -> None:
        cmp = model.by_uuid("b327d900-abd2-4138-a111-9ff0684739d8")
        assert isinstance(cmp, mm.pa.PhysicalComponent)
        func = model.by_uuid("6320d7e0-cfba-4298-a10c-820b02b5c7a9")
        assert isinstance(func, mm.pa.PhysicalFunction)
        previous_ports = len(func.inputs) + len(func.outputs)
        yml = f"""\
            - parent: !uuid {cmp.uuid}
              extend:
                physical_links:
                  - name: My new link
                    allocated_functional_exchanges:
                      - !promise my_exchange
            - parent: !uuid {func.uuid}
              extend:
                exchanges:
                  - source: !promise first-port
                    target: !promise second-port
                    promise_id: my_exchange
            - parent: !uuid {func.uuid}
              extend:
                outputs:
                  - name: First port
                    promise_id: first-port
                inputs:
                  - name: Second port
                    promise_id: second-port
            """

        decl.apply(model, io.StringIO(yml))

        assert len(func.inputs) + len(func.outputs) == previous_ports + 2
        ex = cmp.physical_links.by_name("My new link")
        assert len(ex.allocated_functional_exchanges) == 1
        port = ex.allocated_functional_exchanges[0].source
        assert port is not None
        assert port.name == "First port"


class TestApplyPromises:
    @staticmethod
    def test_promises_can_backwards_reference_objects(
        model: m.MelodyModel,
    ) -> None:
        root_func = model.by_uuid(ROOT_FUNCTION)
        root_comp = model.by_uuid(ROOT_COMPONENT)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the unit test
                    promise_id: pass-test
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise pass-test
            """
        expected_len = len(root_comp.allocated_functions) + 1

        decl.apply(model, io.StringIO(yml))

        actual_len = len(root_comp.allocated_functions)
        assert actual_len == expected_len
        assert "pass the unit test" in root_func.functions.by_name
        assert "pass the unit test" in root_comp.allocated_functions.by_name

    @staticmethod
    def test_promises_can_forward_reference_objects(
        model: m.MelodyModel,
    ) -> None:
        root_func = model.by_uuid(ROOT_FUNCTION)
        root_comp = model.by_uuid(ROOT_COMPONENT)
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise pass-test
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the unit test
                    promise_id: pass-test
            """
        expected_len = len(root_comp.allocated_functions) + 1

        decl.apply(model, io.StringIO(yml))

        actual_len = len(root_comp.allocated_functions)
        assert actual_len == expected_len
        assert "pass the unit test" in root_func.functions.by_name
        assert "pass the unit test" in root_comp.allocated_functions.by_name

    @staticmethod
    @pytest.mark.parametrize(
        "order",
        [
            pytest.param((0, 1), id="backward"),
            pytest.param((1, 0), id="forward"),
        ],
    )
    def test_promises_on_simple_attributes_can_reference_objects(
        model: m.MelodyModel, order
    ) -> None:
        root_func = model.by_uuid(ROOT_FUNCTION)
        snippets = (
            f"""
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                inputs:
                  - name: My input
                    promise_id: inport
                outputs:
                  - name: My output
                    promise_id: outport
            """,
            f"""
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                exchanges:
                  - name: Test exchange
                    source: !promise outport
                    target: !promise inport
            """,
        )
        yml = snippets[order[0]] + snippets[order[1]]

        decl.apply(model, io.StringIO(yml))

        assert "Test exchange" in root_func.exchanges.by_name
        exc = root_func.exchanges.by_name("Test exchange")
        assert exc.source == root_func.outputs.by_name("My output")
        assert exc.target == root_func.inputs.by_name("My input")

    @staticmethod
    def test_reused_promise_ids_cause_an_exception(
        model: m.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the first unit test
                    promise_id: colliding-promise-id
                  - name: pass the second unit test
                    promise_id: colliding-promise-id
            """

        with pytest.raises(ValueError, match=r"\bcolliding-promise-id\b"):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    def test_unfulfilled_promises_raise_an_exception(
        model: m.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise pass-test
            """

        with pytest.raises(decl.UnfulfilledPromisesError, match="^pass-test$"):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    def test_promises_are_resolved_for_parent_objects(
        model: m.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !promise pass-test
              set:
                name: pass the unit test
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - promise_id: pass-test
            - parent: !promise pass-test
              set:
                description: Makes sure that the test passes.
            """

        resolved = decl.apply(model, io.StringIO(yml))

        assert decl.Promise("pass-test") in resolved
        func = resolved[decl.Promise("pass-test")]
        assert isinstance(func, mm.la.LogicalFunction)
        assert func.name == "pass the unit test"
        assert func.description == "Makes sure that the test passes."


class TestApplyModify:
    @staticmethod
    def test_modify_operations_change_model_objects_in_place(
        model: m.MelodyModel,
    ) -> None:
        newname = "Coffee machine"
        root_component = model.by_uuid(ROOT_COMPONENT)
        assert root_component.name != newname
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              set:
                name: {newname}
            """

        decl.apply(model, io.StringIO(yml))

        assert root_component.name == newname

    @staticmethod
    def test_modify_can_set_attributes_to_promises(
        model: m.MelodyModel,
    ) -> None:
        root_component = model.by_uuid(ROOT_COMPONENT)
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              set:
                allocated_functions:
                  - !promise make-coffee
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: make coffee
                    promise_id: make-coffee
            """
        assert "make coffee" not in root_component.allocated_functions.by_name

        decl.apply(model, io.StringIO(yml))

        assert "make coffee" in root_component.allocated_functions.by_name

    @staticmethod
    def test_modify_can_set_attributes_to_uuid_references(
        model: m.MelodyModel,
    ) -> None:
        root_component = model.by_uuid(ROOT_COMPONENT)
        root_function = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              set:
                allocated_functions:
                  - !uuid {ROOT_FUNCTION}
            """
        assert root_function not in root_component.allocated_functions

        decl.apply(model, io.StringIO(yml))

        assert root_function in root_component.allocated_functions

    @staticmethod
    def test_modifying_to_a_list_removes_all_previous_list_members(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              set:
                functions:
                  - name: survive
            """
        assert len(root_function.functions) > 0

        decl.apply(model, io.StringIO(yml))

        assert len(root_function.functions) == 1
        assert root_function.functions[0].name == "survive"


class TestApplyDelete:
    @staticmethod
    def test_delete_operations_remove_child_objects_from_the_model(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        assert len(root_function.functions) > 0
        subfunc: str = root_function.functions[0].uuid
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              delete:
                functions:
                  - !uuid {subfunc}
            """
        assert subfunc in root_function.functions.by_uuid
        expected_len = len(root_function.functions) - 1

        decl.apply(model, io.StringIO(yml))

        assert len(root_function.functions) == expected_len
        assert subfunc not in root_function.functions.by_uuid
        with pytest.raises(KeyError, match=subfunc):
            model.by_uuid(subfunc)

    @staticmethod
    def test_delete_operations_delete_attributes_if_no_list_of_uuids_was_given(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        assert isinstance(root_function, mm.la.LogicalFunction)
        assert len(root_function.functions) > 0
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              delete:
                functions:
            """

        decl.apply(model, io.StringIO(yml))

        assert len(root_function.functions) == 0


class TestApplySync:
    @staticmethod
    def test_sync_operation_without_find_key_raises_an_error(
        model: m.MelodyModel,
    ) -> None:
        yml = f"""\
        - parent: !uuid {ROOT_FUNCTION}
          sync:
            functions:
              - set:
                  name: "The new name"
        """
        with pytest.raises(ValueError, match="find"):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    def test_sync_operation_finds_and_modifies_existing_objects_in_the_list(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        new_description = "This is a function."
        assert root_function.functions[0].description != new_description
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {root_function.functions[0].name}
                    set:
                      description: {new_description!r}
            """

        decl.apply(model, io.StringIO(yml))

        assert root_function.functions[0].description == new_description

    @staticmethod
    def test_sync_operation_honors_type_hints(
        model: m.MelodyModel,
    ) -> None:
        yml = """\
            - parent: !uuid 3c2d312c-37c9-41b5-8c32-67578fa52dc3
              sync:
                attributes:
                  - find:
                      _type: StringValueAttribute
                    promise_id: obj
            """

        promises = decl.apply(model, io.StringIO(yml))

        assert decl.Promise("obj") in promises
        obj = promises[decl.Promise("obj")]
        assert isinstance(obj, reqif.StringValueAttribute)
        assert obj.uuid == "ee8a69ef-61b9-4db9-9a0f-628e5d4704e1"

    @staticmethod
    def test_sync_set_overwrites_iterables(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        subfunc = root_function.functions[0]
        old_port = subfunc.inputs.create(name="Water port")
        new_port_name = "Power port"
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {root_function.functions[0].name}
                    set:
                      inputs:
                        - name: {new_port_name}
            """

        decl.apply(model, io.StringIO(yml))

        assert new_port_name in subfunc.inputs.by_name
        assert old_port not in subfunc.inputs

    @staticmethod
    def test_sync_extend_does_not_overwrite_iterables(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        subfunc = root_function.functions[0]
        old_port = subfunc.inputs.create(name="Water port")
        new_port_name = "Power port"
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {subfunc.name}
                    extend:
                      inputs:
                        - name: {new_port_name}
            """

        decl.apply(model, io.StringIO(yml))

        assert new_port_name in subfunc.inputs.by_name
        assert old_port in subfunc.inputs

    @staticmethod
    def test_sync_operation_recursive(model: m.MelodyModel) -> None:
        root_package = model.la.data_pkg
        assert root_package is not None
        package_name = "The new package"
        subpackage_name = "The new subpackage"
        assert package_name not in root_package.packages.by_name
        yml = f"""\
            - parent: !uuid {root_package.uuid}
              sync:
                packages:
                  - find:
                      name: "{package_name}"
                    sync:
                      packages:
                        - find:
                            name: {subpackage_name}

            """

        decl.apply(model, io.StringIO(yml))
        assert package_name in root_package.packages.by_name
        package = root_package.packages.by_name(package_name)
        assert subpackage_name in package.packages.by_name

    @staticmethod
    def test_sync_does_not_duplicate_elements_during_deferred_nested_syncs(
        model: m.MelodyModel,
    ) -> None:
        yml = """\
            - parent: !find {_type: LogicalArchitecture}
              sync:
                requirement_modules:
                  - find:
                      long_name: Test module
                    promise_id: module
                    sync:
                      requirements:
                        - find:
                            identifier: req
                          promise_id: req
                          set:
                            long_name: Test Requirement
                            type: !promise reqtype
                          sync:
                            attributes:
                              - find:
                                  _type: StringValueAttribute
                                  definition: !promise attrdef
                                set:
                                  value: The value
                requirement_types_folders:
                  - find:
                      long_name: Test types folder
                    sync:
                      types:
                        - find:
                            _type: RequirementType
                            long_name: Test type
                          promise_id: reqtype
                          sync:
                            attributes:
                              - find:
                                  _type: AttributeDefinition
                                  long_name: Attribute
                                promise_id: attrdef
            """

        resolved = decl.apply(model, io.StringIO(yml))

        module = resolved[decl.Promise("module")]
        assert isinstance(module, reqif.CapellaModule)
        assert len(module.requirements) == 1

        req = resolved[decl.Promise("req")]
        assert isinstance(req, reqif.Requirement)
        assert req.type is not None
        assert req.type == resolved[decl.Promise("reqtype")]
        assert len(req.attributes) == 1

    @staticmethod
    def test_sync_operation_creates_a_new_object_if_it_didnt_find_a_match(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        new_name = "The new function"
        new_description = "This is a new function."
        assert new_name not in root_function.functions.by_name
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {new_name}
                    set:
                      description: {new_description!r}
            """

        decl.apply(model, io.StringIO(yml))

        assert new_name in root_function.functions.by_name
        func = root_function.functions.by_name(new_name, single=True)
        assert func.description == new_description

    @staticmethod
    def test_sync_operation_resolves_promises_with_newly_created_objects(
        model: m.MelodyModel,
    ) -> None:
        newname = "The new function"
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {newname}
                    promise_id: my_function
            """

        resolved = decl.apply(model, io.StringIO(yml))

        assert resolved
        assert decl.Promise("my_function") in resolved
        new_func = resolved[decl.Promise("my_function")]
        assert isinstance(new_func, mm.la.LogicalFunction)
        assert new_func.name == newname

    @staticmethod
    def test_sync_operation_resolves_promises_with_existing_objects(
        model: m.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {root_function.functions[0].name}
                    promise_id: my_function
            """

        resolved = decl.apply(model, io.StringIO(yml))

        assert resolved == {
            decl.Promise("my_function"): root_function.functions[0]
        }

    @staticmethod
    def test_sync_operation_can_resolve_multiple_promises_with_one_object(
        model: m.MelodyModel,
    ) -> None:
        function = model.by_uuid(ROOT_FUNCTION).functions[0]
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              sync:
                functions:
                  - find:
                      name: {function.name}
                    promise_id: promise-1
                  - find:
                      name: {function.name}
                    promise_id: promise-2
            """

        resolved = decl.apply(model, io.StringIO(yml))

        assert resolved == {
            decl.Promise("promise-1"): function,
            decl.Promise("promise-2"): function,
        }


class TestStrictMetadata:
    @staticmethod
    def test_strict_apply_requires_nonempty_metadata(model: m.MelodyModel):
        yml = textwrap.dedent(
            """\
            {}
            ---
            []
            """
        )

        with pytest.raises(ValueError, match="^No metadata found"):
            decl.apply(model, io.StringIO(yml), strict=True)

    @staticmethod
    def test_capellambse_version_outdated(model: m.MelodyModel, monkeypatch):
        modelinfo = m.ModelInfo(
            url="git+https://decl-yaml.invalid/ignored-anyway.git",
            title="Testmodel",
            entrypoint=pathlib.PurePosixPath("testmodel.aird"),
            resources={
                "\x00": git.GitHandlerInfo(
                    url="git+https://decl-yaml.invalid/testmodel.git",
                    branch="master",
                    revision="Baseline-1.0",
                    rev_hash="0000000000000000000000000000000000000000",
                )
            },
            capella_version="5.0.0",
            viewpoints={"org.polarsys.capella.core.viewpoint": "5.0.0"},
            diagram_cache=None,
        )
        monkeypatch.setattr(m.MelodyModel, "info", modelinfo)

        yml = textwrap.dedent(
            """\
            written_by:
              capellambse: '99999.99.99'
            model:
              url: 'git+https://decl-yaml.invalid/testmodel.git'
              revision: '0000000000000000000000000000000000000000'
              entrypoint: 'testmodel.aird'
            ---
            []
            """
        )

        with pytest.raises(ValueError, match=r"99999\.99\.99"):
            decl.apply(model, io.StringIO(yml), strict=True)

    @staticmethod
    def test_declared_capellambse_version_malformed(
        model: m.MelodyModel, monkeypatch
    ):
        modelinfo = m.ModelInfo(
            url="git+https://decl-yaml.invalid/ignored-anyway.git",
            title="Testmodel",
            entrypoint=pathlib.PurePosixPath("testmodel.aird"),
            resources={
                "\x00": git.GitHandlerInfo(
                    url="git+https://decl-yaml.invalid/testmodel.git",
                    branch="master",
                    revision="Baseline-1.0",
                    rev_hash="0000000000000000000000000000000000000000",
                )
            },
            capella_version="5.0.0",
            viewpoints={"org.polarsys.capella.core.viewpoint": "5.0.0"},
            diagram_cache=None,
        )
        monkeypatch.setattr(m.MelodyModel, "info", modelinfo)

        yml = textwrap.dedent(
            """\
            written_by:
              capellambse: 'this-is-not-a-pep440-version'
            model:
              url: 'git+https://decl-yaml.invalid/testmodel.git'
              revision: '0000000000000000000000000000000000000000'
              entrypoint: 'testmodel.aird'
            ---
            []
            """
        )

        with pytest.raises(ValueError, match="this-is-not-a-pep440-version"):
            decl.apply(model, io.StringIO(yml), strict=True)

    @staticmethod
    def test_url_mismatch(model: m.MelodyModel, monkeypatch):
        modelinfo = m.ModelInfo(
            url="git+https://decl-yaml.invalid/ignored-anyway.git",
            title="Testmodel",
            entrypoint=pathlib.PurePosixPath("testmodel.aird"),
            resources={
                "\x00": git.GitHandlerInfo(
                    url="git+https://decl-yaml.invalid/testmodel.git",
                    branch="master",
                    revision="Baseline-1.0",
                    rev_hash="0000000000000000000000000000000000000000",
                )
            },
            capella_version="5.0.0",
            viewpoints={"org.polarsys.capella.core.viewpoint": "5.0.0"},
            diagram_cache=None,
        )
        monkeypatch.setattr(m.MelodyModel, "info", modelinfo)

        yml = textwrap.dedent(
            """\
            written_by:
              capellambse: '0.0.1'
            model:
              url: 'git+https://decl-yaml.invalid/other.git'
              revision: '0000000000000000000000000000000000000000'
              entrypoint: 'testmodel.aird'
            ---
            []
            """
        )

        with pytest.raises(ValueError, match=r"/other\.git"):
            decl.apply(model, io.StringIO(yml), strict=True)

    @staticmethod
    def test_entrypoint_mismatch(model: m.MelodyModel, monkeypatch):
        modelinfo = m.ModelInfo(
            url="git+https://decl-yaml.invalid/ignored-anyway.git",
            title="Testmodel",
            entrypoint=pathlib.PurePosixPath("testmodel.aird"),
            resources={
                "\x00": git.GitHandlerInfo(
                    url="git+https://decl-yaml.invalid/testmodel.git",
                    branch="master",
                    revision="Baseline-1.0",
                    rev_hash="0000000000000000000000000000000000000000",
                )
            },
            capella_version="5.0.0",
            viewpoints={"org.polarsys.capella.core.viewpoint": "5.0.0"},
            diagram_cache=None,
        )
        monkeypatch.setattr(m.MelodyModel, "info", modelinfo)

        yml = textwrap.dedent(
            """\
            written_by:
              capellambse: '0.0.1'
            model:
              url: git+https://decl-yaml.invalid/testmodel.git
              revision: '0000000000000000000000000000000000000000'
              entrypoint: othermodel.aird
            ---
            []
            """
        )

        with pytest.raises(ValueError, match=r"othermodel\.aird"):
            decl.apply(model, io.StringIO(yml), strict=True)

    @staticmethod
    def test_model_revision_doesnt_match(model: m.MelodyModel, monkeypatch):
        modelinfo = m.ModelInfo(
            url="git+https://decl-yaml.invalid/ignored-anyway.git",
            title="Testmodel",
            entrypoint=pathlib.PurePosixPath("testmodel.aird"),
            resources={
                "\x00": git.GitHandlerInfo(
                    url="git+https://decl-yaml.invalid/testmodel.git",
                    branch="master",
                    revision="Baseline-1.0",
                    rev_hash="0000000000000000000000000000000000000000",
                )
            },
            capella_version="5.0.0",
            viewpoints={"org.polarsys.capella.core.viewpoint": "5.0.0"},
            diagram_cache=None,
        )
        monkeypatch.setattr(m.MelodyModel, "info", modelinfo)

        yml = textwrap.dedent(
            """\
            written_by:
              capellambse: '0.0.1'
            model:
              url: git+https://decl-yaml.invalid/testmodel.git
              revision: 'ffffffffffffffffffffffffffffffffffffffff'
              entrypoint: testmodel.aird
            ---
            []
            """
        )

        with pytest.raises(ValueError, match=40 * "f"):
            decl.apply(model, io.StringIO(yml), strict=True)


@pytest.mark.parametrize("filename", ["coffee-machine.yml"])
def test_full_example(model: m.MelodyModel, filename: str):
    decl.apply(model, DATAPATH / filename)


def test_cli_applies_a_yaml_and_saves_the_model_back(tmp_path: pathlib.Path):
    shutil.copytree(Models.test7_0, tmp_path / "model")
    model = next(tmp_path.joinpath("model").glob("*.aird"))
    semmodel = model.with_suffix(".capella")
    oldhash = hashlib.sha256(semmodel.read_bytes()).hexdigest()
    declfile = DATAPATH / "coffee-machine.yml"

    cli = subprocess.run(
        [sys.executable, "-mcapellambse.decl", f"--model={model}", declfile],
        cwd=INSTALLED_PACKAGE.parent,
        check=False,
    )

    assert cli.returncode == 0, "CLI process exited unsuccessfully"
    newhash = hashlib.sha256(semmodel.read_bytes()).hexdigest()
    assert newhash != oldhash, "Files on disk didn't change"



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Some additional glue to tie ReqIF into capellambse."""

__all__ = ["init"]

import capellambse.model as m
from capellambse.metamodel import cs

from . import capellarequirements as cr
from . import requirements as rq


def init() -> None:
    m.ModelElement.requirements = cr.ElementRelationAccessor()  # type: ignore[deprecated]
    m.ModelElement.requirements_relations = m.Backref[rq.AbstractRelation](
        (rq.NS, "AbstractRelation"), "source", "target"
    )
    cs.BlockArchitecture.requirement_modules = m.Filter(
        "extensions", (cr.NS, "CapellaModule")
    )
    cs.BlockArchitecture.requirement_types_folders = m.Filter(
        "extensions", (cr.NS, "CapellaTypesFolder")
    )

    cs.BlockArchitecture.all_requirements = property(
        lambda self: self._model.search((rq.NS, "Requirement"), below=self)
    )
    cs.BlockArchitecture.all_requirement_types = property(
        lambda self: self._model.search((rq.NS, "RequirementType"), below=self)
    )
    cs.BlockArchitecture.all_module_types = property(
        lambda self: self._model.search((rq.NS, "ModuleType"), below=self)
    )
    cs.BlockArchitecture.all_relation_types = property(
        lambda self: self._model.search((rq.NS, "RelationType"), below=self)
    )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
import enum

import capellambse.model as m

from . import namespaces as ns

NS = ns.LIBRARIES


@m.stringy_enum
@enum.unique
class AccessPolicy(enum.Enum):
    READ_ONLY = "readOnly"
    READ_AND_WRITE = "readAndWrite"


class LibraryAbstractElement(m.ModelElement, abstract=True):
    pass


class ModelInformation(LibraryAbstractElement):
    references = m.Containment["LibraryReference"](
        "ownedReferences", (NS, "LibraryReference")
    )
    version = m.Association["ModelVersion"]((NS, "ModelVersion"), "version")


class LibraryReference(LibraryAbstractElement):
    library = m.Single["ModelInformation"](
        m.Association((NS, "ModelInformation"), "library")
    )
    access_policy = m.EnumPOD("accessPolicy", AccessPolicy)
    version = m.Association["ModelVersion"]((NS, "ModelVersion"), "version")


class ModelVersion(LibraryAbstractElement):
    major_version_number = m.IntPOD("majorVersionNumber")
    minor_version_number = m.IntPOD("minorVersionNumber")
    last_modified_file_stamp = m.IntPOD("lastModifiedFileStamp")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The module provides management and evaluation of validation rules.

Validation rules are conditions ensuring that specific modeling
guidelines are followed. These rules apply to particular types of model
elements or to diagrams, and values of metrics. By evaluating each rule,
the module generates validation results, indicating whether the
corresponding guideline has been satisfied or not. This way, the module
helps maintain the quality and consistency of the model.
"""

from ._validate import *

from . import rules as rules  # isort: skip

import capellambse
import capellambse.model as m
from capellambse.metamodel import cs


def init() -> None:
    capellambse.MelodyModel.validation = property(  # type: ignore[attr-defined]
        ModelValidation
    )
    capellambse.MelodyModel.validate = property(  # type: ignore[attr-defined]
        lambda self: self.validation.validate
    )

    m.ModelElement.validation = m.AlternateAccessor(ElementValidation)
    m.ModelElement.validate = property(lambda self: self.validation.validate)
    cs.ComponentArchitecture.validation = m.AlternateAccessor(LayerValidation)
